from typing import Tuple
import gzip
import logging
import os
import sys
from os import path

import numpy as np
import pandas as pd
from bean import Allele, ReporterScreen
from Bio import SeqIO
from Bio.SeqIO.QualityIO import FastqPhredIterator
from tqdm import tqdm

from ._supporting_fn import (
    _base_edit_to_from,
    _check_readname_match,
    _get_edited_allele_crispresso,
    _get_fastq_handle,
    _multiindex_dict_to_df,
    _read_count_match,
    _read_is_good_quality,
    _write_alignment_matrix,
    _write_paired_end_reads,
    revcomp,
)

logging.basicConfig(
    level=logging.INFO,
    format="%(levelname)-5s @ %(asctime)s:\n\t %(message)s \n",
    datefmt="%a, %d %b %Y %H:%M:%S",
    stream=sys.stderr,
    filemode="w",
)
error = logging.critical
warn = logging.warning
debug = logging.debug
info = logging.info


class InputFileError(Exception):
    """Raised when the input file is not valid."""


class NoReadsAfterQualityFiltering(Exception):
    """Raised when no reads are remaining after quality filtering."""


strand_str_to_int = {"neg": -1, "pos": 1, "-": -1, "+": 1}


def _get_stranded_guide_offset(strand: int, start_pos: int, guide_len: int) -> int:
    if strand == -1:
        offset = start_pos + 32 - 6 - 1
    elif strand == 1:
        offset = start_pos - (32 - 6 - guide_len)
    return offset


class GuideEditCounter:
    def __init__(self, **kwargs):
        self.R1_filename = kwargs["R1"]
        self.R2_filename = kwargs["R2"]
        self.base_edited_from = kwargs["edited_base"]
        self.base_edited_to = _base_edit_to_from(self.base_edited_from)
        self.min_average_read_quality = kwargs["min_average_read_quality"]
        self.min_single_bp_quality = kwargs["min_single_bp_quality"]

        self.guides_info_df = pd.read_csv(kwargs["sgRNA_filename"])
        self.guides_has_strands = (
            "strand" in self.guides_info_df.columns
        )  # @TODO: fix this?
        if self.guides_has_strands:
            info("Considering `strand` column in sgRNA info file")
            if "start_pos" not in self.guides_info_df.columns:
                raise InputFileError("Guide `start_pos` not in sgRNA info file.")
            if "guide_len" not in self.guides_info_df.columns:
                self.guides_info_df["guide_len"] = self.guides_info_df.sequence.map(len)
        else:
            info("Ignoring guide strands, all guides are considered positive")

        self.qstart_R1 = kwargs["qstart_R1"]
        self.qend_R1 = kwargs["qend_R1"]
        self.qstart_R2 = kwargs["qstart_R2"]
        self.qend_R2 = kwargs["qend_R2"]
        self.name = kwargs["name"]

        self.count_only_bcmatched = False
        self.sgRNA_filename = kwargs["sgRNA_filename"]
        self._set_sgRNA_df()

        self.database_id = self._get_database_name()
        self.output_dir = os.path.join(
            os.path.abspath(kwargs["output_folder"]),
            f"bean_count_{self.database_id}",
        )
        self._write_start_log()

        self.screen = ReporterScreen(
            X=np.zeros((len(self.guides_info_df), 1)),
            X_edit=np.zeros((len(self.guides_info_df), 1)),
            X_bcmatch=np.zeros((len(self.guides_info_df), 1)),
            guides=self.guides_info_df,
            samples=pd.DataFrame(index=[self.database_id]),
            target_base_change=f"{self.base_edited_from}>{self.base_edited_to}",
            tiling=kwargs["tiling"],
        )
        self.count_guide_edits = kwargs["count_guide_edits"]
        if self.count_guide_edits:
            self.screen.uns["guide_edit_counts"] = {}
        self.count_reporter_edits = (
            kwargs["count_reporter"] or kwargs["count_guide_reporter_alleles"]
        )
        if self.count_reporter_edits:
            self.screen.uns["edit_counts"] = {}
            self.gstart_reporter = kwargs["gstart_reporter"]
            self.screen.uns["allele_counts"] = {}
        self.count_guide_reporter_alleles = kwargs["count_guide_reporter_alleles"]
        if self.count_guide_reporter_alleles:
            self.guide_to_guide_reporter_allele = {}
        self.align_score_threshold = 80
        self.target_pos_col = kwargs["target_pos_col"]

        self.guide_start_seq = kwargs["guide_start_seq"]
        self.guide_end_seq = kwargs["guide_end_seq"]
        if not self.guide_start_seq == "":
            info(
                f"{self.name}: Using guide_start_seq={self.guide_start_seq} for {self.output_dir}"
            )
        assert (
            self.guide_start_seq == "" or self.guide_end_seq == ""
        ), "Doesn't support both start & end seq matching"
        self.guide_bc = kwargs["guide_bc"]
        if self.guide_bc:
            self.guide_bc_len = kwargs["guide_bc_len"]

        self.offset = kwargs["offset"]
        if self.count_reporter_edits:
            self.reporter_length = kwargs["reporter_length"]
        self.guide_to_allele = {}
        self.n_total_reads = _read_count_match(self.R1_filename, self.R2_filename)

        self.objectify_allele = not kwargs["string_allele"]
        if not self.objectify_allele:
            info(f"{self.name}: Storing allele as strings.")
        self.keep_intermediate = kwargs["keep_intermediate"]
        self.semimatch = 0
        self.bcmatch = 0
        self.nomatch = 0
        self.duplicate_match = 0
        self.duplicate_match_wo_barcode = 0

    def masked_equal(self, seq1, seq2):
        """Tests if two sequences are equal, ignoring the allowed base transition."""
        return seq1.replace(self.base_edited_from, self.base_edited_to) == seq2.replace(
            self.base_edited_from, self.base_edited_to
        )

    def _set_sgRNA_df(self):
        """set gRNA info dataframe"""
        with open(self.sgRNA_filename) as infile:
            sgRNA_df = pd.read_csv(infile)
            if "name" not in sgRNA_df.columns or "sequence" not in sgRNA_df.columns:
                raise InputFileError(
                    "Input gRNA info file doesn't have the column 'name' or 'sequence'."
                )
            if self.count_only_bcmatched and "barcode" not in sgRNA_df.columns:
                raise InputFileError(
                    "Input gRNA info file doesn't have the column 'barcode'."
                )
            sgRNA_df = sgRNA_df.set_index("name")
        self.guides_info_df = sgRNA_df
        self.guides_info_df["masked_sequence"] = self.guides_info_df.sequence.map(
            lambda s: s.replace(self.base_edited_from, self.base_edited_to)
        )
        self.guides_info_df["masked_barcode"] = self.guides_info_df.barcode.map(
            lambda s: s.replace(self.base_edited_from, self.base_edited_to)
        )
        self.guide_lengths = sgRNA_df.sequence.map(lambda s: len(s)).unique()

    def check_filter_fastq(self):
        """Checks if the quality filtered fastq files already exists,
        and use them if the do."""
        self.filtered_R1_filename = self._jp(
            os.path.basename(self.R1_filename).replace(".fastq", "").replace(".gz", "")
            + "_filtered.fastq.gz"
        )
        self.filtered_R2_filename = self._jp(
            os.path.basename(self.R2_filename).replace(".fastq", "").replace(".gz", "")
            + "_filtered.fastq.gz"
        )
        if path.exists(self.filtered_R1_filename) and path.exists(
            self.filtered_R2_filename
        ):
            warn("Using preexisting filtered file")
        else:
            self._check_names_filter_fastq()

    def get_counts(self):
        # infile_R1 = _get_fastq_handle(self.R1_filename)
        # infile_R2 = _get_fastq_handle(self.R2_filename)

        self.nomatch_R1_filename = self.R1_filename.replace(".fastq", "_nomatch.fastq")
        self.nomatch_R2_filename = self.R2_filename.replace(".fastq", "_nomatch.fastq")
        self.semimatch_R1_filename = self.R1_filename.replace(
            ".fastq", "_semimatch.fastq"
        )
        self.semimatch_R2_filename = self.R2_filename.replace(
            ".fastq", "_semimatch.fastq"
        )
        if self.count_reporter_edits:
            _write_alignment_matrix(
                self.base_edited_from,
                self.base_edited_to,
                self.output_dir + "/.aln_mat.txt",
            )

        if self.count_only_bcmatched:
            # count X
            self._get_guide_counts_bcmatch()
        else:  # count both bc matched & unmatched guide counts
            self._get_guide_counts_bcmatch_semimatch()

        if self.count_guide_edits:
            self.screen.uns["guide_edit_counts"] = _multiindex_dict_to_df(
                self.screen.uns["guide_edit_counts"], "edit", self.database_id
            )
            self.screen.uns["guide_edit_counts"].guide = self.screen.guides.index[
                self.screen.uns["guide_edit_counts"].guide.to_numpy(dtype=int)
            ]

        if self.count_reporter_edits:
            self.screen.uns["edit_counts"] = pd.DataFrame.from_dict(
                self.screen.uns["edit_counts"]
            )

        if self.count_reporter_edits:
            self._write_reporter_alleles()
        if self.count_guide_reporter_alleles:
            self._write_guide_reporter_alleles()
        count_stat_path = self._jp("mapping_stats.txt")
        count_stat_file = open(count_stat_path, "w")
        count_stat_file.write("Read count with \n")
        count_stat_file.write(
            f"Unique guide match without barcode:\t{self.semimatch}\n"
        )
        count_stat_file.write(f"Unique guide match with barcode:\t{self.bcmatch}\n")
        count_stat_file.write(f"No match:\t{self.nomatch}\n")
        count_stat_file.write(
            f"Duplicate match with barcode:\t{self.duplicate_match}\n"
        )
        count_stat_file.write(
            f"No match with barcode & Duplicate match w/o barcode:\t{self.duplicate_match_wo_barcode}\n"
        )

    def _write_guide_reporter_alleles(self):
        guides = []
        guide_alleles = []
        reporter_alleles = []
        counts = []
        for guide, allele_to_count in self.guide_to_guide_reporter_allele.items():
            if len(allele_to_count.keys()) == 0:
                continue
            guides.extend([guide] * len(allele_to_count.keys()))
            for key, count in allele_to_count.items():
                reporter_allele, guide_allele = key
                guide_alleles.append(guide_allele)
                reporter_alleles.append(reporter_allele)
                counts.append(count)
        if not (
            len(guides) == len(reporter_alleles) == len(guide_alleles) == len(counts)
        ):
            raise ValueError(
                f"Guides:{len(guides)}, guide_alleles:{len(guide_alleles)}, reporter_alleles: {len(reporter_alleles)}, counts:{len(counts)}"
            )
        self.screen.uns["guide_reporter_allele_counts"] = pd.DataFrame(
            {
                "guide": guides,
                "guide_allele": guide_alleles,
                "reporter_allele": reporter_alleles,
                self.database_id: counts,
            }
        )

        if "guide" in self.screen.uns["guide_reporter_allele_counts"].columns:
            self.screen.uns[
                "guide_reporter_allele_counts"
            ].guide = self.screen.guides.index[
                self.screen.uns["guide_reporter_allele_counts"].guide
            ]

    def _write_reporter_alleles(self):
        guides = []
        alleles = []
        counts = []
        for guide, allele_to_count in self.guide_to_allele.items():
            if len(allele_to_count.keys()) == 0:
                continue
            guides.extend([guide] * len(allele_to_count.keys()))
            for allele, count in allele_to_count.items():
                alleles.append(allele)
                counts.append(count)
        if not (len(guides) == len(alleles) == len(counts)):
            raise ValueError(
                f"Guides:{len(guides)}, alleles:{len(alleles)}, counts:{len(counts)}"
            )
        self.screen.uns["allele_counts"] = pd.DataFrame(
            {"guide": guides, "allele": alleles, self.database_id: counts}
        )

        if "guide" in self.screen.uns["allele_counts"].columns:
            self.screen.uns["allele_counts"].guide = self.screen.guides.index[
                self.screen.uns["allele_counts"].guide
            ]

    def _get_guide_counts_bcmatch(self):
        NotImplemented

    def _count_guide_edits(
        self, matched_guide_idx, R1_record: SeqIO.SeqRecord, single_base_qual_cutoff=30
    ):
        if self.guides_has_strands:
            strand = self.screen.guides.strand[matched_guide_idx]
            guide_strand = strand_str_to_int.get(strand, 1)
        else:
            guide_strand = 1
        ref_guide_seq = self.screen.guides.sequence[matched_guide_idx]
        read_guide_seq, read_guide_qual = self.get_guide_seq_qual(
            R1_record, len(ref_guide_seq)
        )
        guide_edit_allele, score = _get_edited_allele_crispresso(
            ref_seq=ref_guide_seq,
            query_seq=read_guide_seq,
            ref_base=self.base_edited_from,
            alt_base=self.base_edited_to,
            aln_mat_path=self.output_dir + "/.aln_mat.txt",
            offset=0,
            strand=guide_strand,
            start_pos=0,
            end_pos=len(ref_guide_seq),
            positionwise_quality=np.array(read_guide_qual),
            quality_thres=single_base_qual_cutoff,
            objectify_allele=self.objectify_allele,
        )
        return (guide_edit_allele, score)

    def _get_strand_offset_from_guide_index(self, guide_idx: int) -> Tuple[int, int]:
        """Returns guide starnd and offset for a given guide index."""
        if self.guides_has_strands:
            strand = self.screen.guides.strand[guide_idx]
            if strand in strand_str_to_int:
                guide_strand = strand_str_to_int[strand]
                offset = _get_stranded_guide_offset(
                    strand=guide_strand,
                    start_pos=self.screen.guides.start_pos[guide_idx],
                    guide_len=self.screen.guides.guide_len[guide_idx],
                )
            else:
                guide_strand = 1
                offset = 0

        else:
            guide_strand = 1
            if self.target_pos_col in self.screen.guides.columns:
                offset = -(self.screen.guides[self.target_pos_col][guide_idx] - 1)
            else:
                offset = 0
        return (guide_strand, offset)

    def _update_counted_allele(self, guide_idx: int, allele: Allele) -> None:
        """Add allele count to self.guide_to_allele dictionary."""
        if guide_idx in self.guide_to_allele.keys():
            if allele in self.guide_to_allele[guide_idx].keys():
                self.guide_to_allele[guide_idx][allele] += 1
            else:
                self.guide_to_allele[guide_idx][allele] = 1
        else:
            self.guide_to_allele[guide_idx] = {allele: 1}

    def _update_counted_allele_and_guideAllele(
        self, guide_idx: int, allele: Allele, guide_allele: Allele
    ) -> None:
        """Add count of (guide allele, reporter allele) combination to self.guide_reporter_allele dictionary."""
        if guide_idx in self.guide_to_guide_reporter_allele.keys():
            if (allele, guide_allele) in self.guide_to_guide_reporter_allele[
                guide_idx
            ].keys():
                self.guide_to_guide_reporter_allele[guide_idx][
                    (allele, guide_allele)
                ] += 1
            else:
                self.guide_to_guide_reporter_allele[guide_idx][
                    (allele, guide_allele)
                ] = 1
        else:
            self.guide_to_guide_reporter_allele[guide_idx] = {(allele, guide_allele): 1}

    def _count_reporter_edits(
        self,
        matched_guide_idx: int,
        R1_seq: str,
        R2_record: SeqIO.SeqRecord,
        single_base_qual_cutoff: str = 30,
        guide_allele: Allele = None,
    ):
        """
        Count edits in a single read to save as allele.

        Args
        --
        matched_guide_idx: index of guides in self.screen.guides to get information from
        R1_seq: Read1 sequence
        R2_record: Read2 sequence record with quality
        single_base_qual_cutoff: Ignore this base if the Phread quality score is less than this threshold
        guide_allele: Allele from baseedit in gRNA spacer sequence when paired with guide allele.
        """
        ref_reporter_seq = self.screen.guides.reporter[matched_guide_idx]
        read_reporter_seq, read_reporter_qual = self.get_reporter_seq_qual(R2_record)

        guide_strand, offset = self._get_strand_offset_from_guide_index(
            matched_guide_idx
        )

        allele, score = _get_edited_allele_crispresso(
            ref_seq=ref_reporter_seq,
            query_seq=read_reporter_seq,
            ref_base=self.base_edited_from,
            alt_base=self.base_edited_to,
            aln_mat_path=self.output_dir + "/.aln_mat.txt",
            offset=offset,
            strand=guide_strand,
            positionwise_quality=np.array(read_reporter_qual),
            quality_thres=single_base_qual_cutoff,
            objectify_allele=self.objectify_allele,
        )

        if score < self.align_score_threshold:
            self.semimatch += 1
            self.bcmatch -= 1
            return

        if self.count_reporter_edits:
            self._update_counted_allele(matched_guide_idx, allele)

        if self.count_guide_reporter_alleles and (guide_allele is not None):
            self._update_counted_allele_and_guideAllele(
                matched_guide_idx, allele, guide_allele
            )

    def _get_guide_counts_bcmatch_semimatch(
        self, bcmatch_layer="X_bcmatch", semimatch_layer="X"
    ):
        self.screen.layers[semimatch_layer] = np.zeros_like((self.screen.X))
        R1_iter, R2_iter = self._get_fastq_iterators()
        if self.keep_intermediate:
            outfile_R1_nomatch, outfile_R2_nomatch = self._get_fastq_handle("nomatch")
            outfile_R1_semimatch, outfile_R2_semimatch = self._get_fastq_handle(
                "semimatch"
            )
            outfile_R1_dup_wo_bc, outfile_R2_dup_wo_bc = self._get_fastq_handle(
                "duplicate_wo_barcode"
            )
            outfile_R1_dup, outfile_R2_dup = self._get_fastq_handle("duplicate")
        with tqdm(
            enumerate(zip(R1_iter, R2_iter)),
            total=self.n_reads_after_filtering,
            postfix=f"n_read={self.bcmatch}",
        ) as tqdm_reads:
            for i, (r1, r2) in tqdm_reads:
                R1_seq = str(r1.seq)
                R2_seq = str(r2.seq)

                bc_match, semimatch = self._match_read_to_sgRNA_bcmatch_semimatch(
                    R1_seq, R2_seq
                )

                if len(bc_match) == 0:
                    if len(semimatch) == 0:  # no guide match
                        if self.keep_intermediate:
                            _write_paired_end_reads(
                                r1, r2, outfile_R1_nomatch, outfile_R2_nomatch
                            )
                        self.nomatch += 1
                    elif len(semimatch) >= 2:  # Duplicate match if w/o barcode
                        if self.keep_intermediate:
                            _write_paired_end_reads(
                                r1, r2, outfile_R1_dup_wo_bc, outfile_R2_dup_wo_bc
                            )
                        self.duplicate_match_wo_barcode += 1
                    else:  # guide match with no barcode match
                        matched_guide_idx = semimatch[0]
                        self.screen.layers[semimatch_layer][matched_guide_idx, 0] += 1
                        if self.count_guide_edits:
                            self._count_guide_edits(matched_guide_idx, r1)
                        self.semimatch += 1

                elif len(bc_match) >= 2:  # duplicate mapping
                    if self.keep_intermediate:
                        _write_paired_end_reads(r1, r2, outfile_R1_dup, outfile_R2_dup)
                    self.duplicate_match += 1

                else:  # unique barcode match
                    matched_guide_idx = bc_match[0]
                    self.screen.layers[bcmatch_layer][matched_guide_idx, 0] += 1
                    self.bcmatch += 1
                    if self.count_guide_edits or self.count_guide_reporter_alleles:
                        guide_allele, _ = self._count_guide_edits(matched_guide_idx, r1)
                    if self.count_reporter_edits:
                        # TODO: what if reporter seq doesn't match barcode & guide?
                        if self.count_guide_reporter_alleles:
                            self._count_reporter_edits(
                                matched_guide_idx, R1_seq, r2, guide_allele=guide_allele
                            )
                        else:
                            self._count_reporter_edits(matched_guide_idx, R1_seq, r2)
                tqdm_reads.postfix = f"n_read={self.bcmatch}"
                tqdm_reads.update()

        self.screen.X = (
            self.screen.layers[semimatch_layer] + self.screen.layers[bcmatch_layer]
        )

    def _write_allele(self, guide_idx: int, allele: Allele, uns_key="allele_counts"):
        if (guide_idx, str(allele)) in self.screen.uns[uns_key].keys():
            self.screen.uns[uns_key][(guide_idx, str(allele))] += 1
        else:
            self.screen.uns[uns_key][(guide_idx, str(allele))] = 1

    def _write_edits(self, guide_idx: int, allele: Allele, uns_key="edit_counts"):
        for edit in allele.edits:
            if (guide_idx, str(edit)) in self.screen.uns[uns_key].keys():
                self.screen.uns[uns_key][(guide_idx, str(edit))] += 1
            else:
                self.screen.uns[uns_key][(guide_idx, str(edit))] = 1

    def _write_guide_reporter_allele(
        self,
        guide_idx: int,
        allele: Allele,
        guide_allele: Allele,
        uns_key="guide_reporter_allele_counts",
    ):
        if len(allele.edits) == 0 and len(guide_allele.edits) == 0:
            return
        if (guide_idx, str(allele), str(guide_allele)) in self.screen.uns[
            uns_key
        ].keys():
            self.screen.uns[uns_key][(guide_idx, str(allele), str(guide_allele))] += 1
        else:
            self.screen.uns[uns_key][(guide_idx, str(allele), str(guide_allele))] = 1

    def get_guide_seq(self, R1_seq, R2_seq, guide_length):
        """This can be edited by user based on the read construct."""
        # _seq_match = np.where(seq.replace(self.base_edited_from, self.base_edited_to) == self.screen.guides.masked_sequence)[0]
        if self.guide_end_seq == "":
            guide_start_idx = R1_seq.replace(
                self.base_edited_from, self.base_edited_to
            ).find(
                self.guide_start_seq.replace(self.base_edited_from, self.base_edited_to)
            )
            if guide_start_idx == -1:
                return None
            if guide_start_idx + guide_length >= len(R1_seq):
                return None
            guide_start_idx = guide_start_idx + len(self.guide_start_seq)
            gRNA_seq = R1_seq[guide_start_idx : (guide_start_idx + guide_length)]
        else:
            guide_end_idx = R1_seq.replace(
                self.base_edited_from, self.base_edited_to
            ).find(
                self.guide_end_seq.replace(self.base_edited_from, self.base_edited_to)
            )
            if guide_end_idx == -1:
                return None
            if guide_end_idx - guide_length < 0:
                return None
            gRNA_seq = R1_seq[(guide_end_idx - guide_length) : guide_end_idx]

        return None if len(gRNA_seq) != guide_length else gRNA_seq

    def get_guide_seq_qual(self, R1_record: SeqIO.SeqRecord, guide_length):
        R1_seq = R1_record.seq
        guide_start_idx = R1_seq.replace(
            self.base_edited_from, self.base_edited_to
        ).find(self.guide_start_seq.replace(self.base_edited_from, self.base_edited_to))
        if guide_start_idx == -1:
            return None, None
        if guide_start_idx + guide_length >= len(R1_seq):
            return None, None
        guide_start_idx = guide_start_idx + len(self.guide_start_seq)
        seq = R1_record[guide_start_idx : (guide_start_idx + guide_length)]
        return (str(seq.seq), seq.letter_annotations["phred_quality"])

    def get_reporter_seq(self, R1_seq, R2_seq):
        """This can be edited by user based on the read construct."""
        return revcomp(
            R2_seq[self.guide_bc_len : (self.guide_bc_len + self.reporter_length)]
        )

    def get_reporter_seq_qual(self, R2_record: SeqIO.SeqRecord):
        seq = R2_record[
            self.guide_bc_len : (self.guide_bc_len + self.reporter_length)
        ].reverse_complement()
        return (str(seq.seq), seq.letter_annotations["phred_quality"])

    def get_barcode(self, R1_seq, R2_seq):
        """This can be edited by user based on the read construct."""
        return revcomp(R2_seq[: self.guide_bc_len])

    def _match_read_to_sgRNA_bcmatch_semimatch(self, R1_seq: str, R2_seq: str):
        # This should be adjusted for each experimental recipes.
        guide_barcode = self.get_barcode(R1_seq, R2_seq)
        bc_match_idx = np.array([])
        semimatch_idx = np.array([])
        for guide_length in self.guide_lengths:
            seq = self.get_guide_seq(R1_seq, R2_seq, guide_length)
            if seq is None:
                continue

            _seq_match = np.where(
                seq.replace(self.base_edited_from, self.base_edited_to)
                == self.screen.guides.masked_sequence
            )[0]
            _bc_match = np.where(
                guide_barcode.replace(self.base_edited_from, self.base_edited_to)
                == self.screen.guides.masked_barcode
            )[0]

            bc_match_idx = np.append(
                bc_match_idx, np.intersect1d(_seq_match, _bc_match)
            )
            semimatch_idx = np.append(semimatch_idx, _seq_match)

        return (bc_match_idx.astype(int), semimatch_idx.astype(int))

    def _get_guide_position_seq_of_read(self, seq):
        guide_start_idx = self._get_guide_start_idx(seq)
        if guide_start_idx == -1:
            return None

        return [
            seq[guide_start_idx : (guide_start_idx + guide_length)]
            for guide_length in self.guide_lengths
        ]

    def _get_guide_start_idx(self, seq):
        start_seq_idx = seq.find(self.guide_start_seq)
        return -1 if start_seq_idx == -1 else start_seq_idx + len(self.guide_start_seq)

    def get_gRNA_barcode(self, R1_seq, R2_seq):
        # This can be adjusted for different construct design.
        return revcomp(R2_seq[: self.guide_bc_len])

    def _get_fastq_handle(
        self,
        out_type: str = None,
    ):
        assert out_type in {
            "semimatch",
            "nomatch",
            "duplicate_wo_barcode",
            "duplicate",
        }
        R1_filename = self.R1_filename.replace(".fastq", f"_{out_type}.fastq")
        R2_filename = self.R2_filename.replace(".fastq", f"_{out_type}.fastq")
        R1_handle = _get_fastq_handle(R1_filename, "w")
        R2_handle = _get_fastq_handle(R2_filename, "w")

        return (R1_handle, R2_handle)

    def _get_fastq_iterators(self):
        R1_handle = _get_fastq_handle(self.R1_filename)
        R2_handle = _get_fastq_handle(self.R2_filename)

        R1_iterator = FastqPhredIterator(R1_handle)
        R2_iterator = FastqPhredIterator(R2_handle)

        return (R1_iterator, R2_iterator)

    def _get_seq_records(self):
        R1_handle = _get_fastq_handle(self.R1_filename)
        R2_handle = _get_fastq_handle(self.R2_filename)
        R1 = list(SeqIO.parse(R1_handle, "fastq"))
        R2 = list(SeqIO.parse(R2_handle, "fastq"))
        R1_handle.close()
        R2_handle.close()
        return (R1, R2)

    def _check_names_filter_fastq(self, filter_by_qual=False):
        if self.min_average_read_quality > 0 or self.min_single_bp_quality > 0:
            info(
                f"Filtering reads with average bp quality < {self.min_average_read_quality} and single bp quality < {self.min_single_bp_quality} ..."
            )

        if self.qend_R1 > 0 or self.qend_R2 > 0:
            info(
                f"In the filtering, bases up to position {self.qend_R1} of R1 and {self.qend_R2} of R2 are considered."
            )

        R1, R2 = self._get_seq_records()

        _check_readname_match(R1, R2)
        if filter_by_qual:
            self.n_reads_after_filtering = self._filter_read_quality(R1, R2)
            if self.n_reads_after_filtering == 0:
                raise NoReadsAfterQualityFiltering(
                    "No reads in input or no reads survived the average or single bp quality filtering."
                )
            else:
                info(
                    "Number of reads in input:%d\tNumber of reads after filtering:%d\n"
                    % (self.n_total_reads, self.n_reads_after_filtering)
                )
        else:
            self.n_reads_after_filtering = self.n_total_reads

    def _filter_read_quality(self, R1=None, R2=None) -> int:
        R1_filtered = gzip.open(self.filtered_R1_filename, "w+")
        R2_filtered = gzip.open(self.filtered_R2_filename, "w+")

        if R1 is None or R2 is None:
            R1, R2 = self._get_seq_records()

        n_reads_after_filtering = 0
        for i, R1_record in enumerate(R1):
            R2_record = R2[i]

            R1_quality_pass = _read_is_good_quality(
                R1_record,
                self.min_average_read_quality,
                self.min_single_bp_quality,
                self.qend_R1,
            )
            R2_quality_pass = _read_is_good_quality(
                R2_record,
                self.min_average_read_quality,
                self.min_single_bp_quality,
                self.qend_R2,
            )

            if R1_quality_pass and R2_quality_pass:
                n_reads_after_filtering += 1
                R1_filtered.write(R1.format("fastq"))
                R2_filtered.write(R2.format("fastq"))
        return n_reads_after_filtering

    def _write_start_log(self):
        try:
            os.makedirs(self.output_dir)
            info(f"Creating Folder {self.output_dir}")
        except OSError:
            info(f"Folder {self.output_dir} already exists.")
        self.log_filename = self._jp("beanCount_RUNNING_LOG.txt")
        logging.getLogger().addHandler(logging.FileHandler(self.log_filename))
        with open(self.log_filename, "w+") as outfile:
            outfile.write(
                "[Command used]:\nCRISPRessoCount %s\n\n[Execution log]:\n"
                % " ".join(sys.argv)
            )

    def _jp(self, filename):
        return os.path.join(self.output_dir, filename)

    def _get_database_name(self):
        get_name_from_fasta = (
            lambda x: os.path.basename(x)
            .replace(".fastq", "")
            .replace(".gz", "")
            .replace("_R1", "")
        )

        return self.name or f"{get_name_from_fasta(self.R1_filename)}"



import numpy as np
import torch
from scipy.optimize import curve_fit

"""
Returns fitted alpha 0 (sum of concentration) of Dirichlet-Multinomial
For the notation, see DESeq paper
"""


def get_valid_idx(x: torch.Tensor, y: torch.Tensor):
    return ~(torch.isnan(x) | torch.isnan(y) | torch.isinf(x) | torch.isinf(y))


def get_valid_vals(x, y):
    valid = get_valid_idx(x, y)
    return (x[valid].detach().cpu().numpy(), y[valid].detach().cpu().numpy())


def get_q(X, sample_size_factors, sample_mask=None):
    """
    Obtain depth-normalized sample mean
    """
    n_reps, n_condits, n_guides, n_alleles = X.shape
    if sample_mask is None:
        sample_mask = torch.ones((n_reps, n_condits))
    q_condit_guide = (
        (X / sample_size_factors[:, :, None, None]) * sample_mask[:, :, None, None]
    ).sum(axis=0) / sample_mask.sum(axis=0)[:, None, None]
    assert q_condit_guide.shape == (
        n_condits,
        n_guides,
        n_alleles,
    ), q_condit_guide.shape
    return q_condit_guide[0, ...]


def get_w(X, sample_size_factors, sample_mask=None):
    """
    Obtain depth-normalized sample variance
    """
    n_reps, n_condits, n_guides, n_alleles = X.shape
    if sample_mask is None:
        sample_mask = torch.ones((n_reps, n_condits), device="cpu")
    q = get_q(X, sample_size_factors, sample_mask)
    se = (X / sample_size_factors[:, :, None, None] - q) ** 2
    w = (se * sample_mask[:, :, None, None]).sum(axis=0) / (sample_mask.sum(axis=0))[
        :, None, None
    ]
    assert w.shape == (n_condits, n_guides, n_alleles), w.shape
    return w[0, ...]


def get_a0(q, w):
    n = q.sum(-1)
    p = q / q.sum(axis=-1)[:, None]
    a0 = torch.nanmean(
        (n[:, None] / ((w - q) / (n[:, None] * p * (1 - p)) + (p / (1 - p)))), axis=-1
    )
    return a0


def linear(x, b0, b1):
    return b0 + b1 * x


def estimate_variance(y, y_est):
    return (y - y_est) ** 2 / len(y)


def shrink_normal_normal(y, y_est, shrink_prior_var: float = 1.0):
    var = estimate_variance(y, y_est)
    prior_weight = var / (var + shrink_prior_var)
    data_weight = 1 - prior_weight
    post_est = prior_weight * y_est + data_weight * y
    return post_est


def get_fitted_alpha0(
    X: torch.Tensor,
    sample_size_factors: torch.Tensor,
    sample_mask: torch.Tensor = None,
    fit: bool = True,
    fit_quantile: float = None,
    shrink: bool = False,
    shrink_prior_var: float = 1.0,
) -> torch.Tensor:  # TODO: set up for variable bin size of X? or scenario when alpha is strained to 0
    """Fits sum of concentration of DirichletMultinomial distribution.

    Args:
        X: allele counts tensor with replicate in the first dimension and allele in the last dimension.
        fit_quantile: if not None, alpha is fitted conservatively with lowest `fit_quantile`
            guides.
        shrink: Return shrinked alpha based on fitted trend, instead of fitted value itself.
    """
    n_reps, n_condits, n_guides, n_alleles = X.shape
    if sample_mask is None:
        sample_mask = torch.ones((n_reps, n_condits), device="cpu")
    w = get_w(X + 1, sample_size_factors, sample_mask=sample_mask)
    q = get_q(X + 1, sample_size_factors, sample_mask=sample_mask)
    n = q.sum(-1)
    a0 = get_a0(q, w)
    assert a0.shape == (n_guides,), a0.shape
    if not fit:
        return (a0, np.nan)

    x, y = get_valid_vals(n.log(), a0.log())
    popt, pcov = curve_fit(linear, x, y)
    print("Linear fit of log(pi_a0) ~ log(q): [b0, b1]={}, cov={}".format(popt, pcov))
    if not fit_quantile is None:
        print(f"Using lowest {fit_quantile} residual to fit again")
        sel_idx = np.where(
            (y - linear(x, *popt)) < np.quantile((y - linear(x, *popt)), fit_quantile)
        )[0]
        if len(sel_idx) < 5:
            print(
                "Too few points available for conservative re-fitting. Consider increasing `fit_quantile`."
            )
        else:
            popt, pcov = curve_fit(linear, x[sel_idx], y[sel_idx])
            print(
                "Adjusted conservative linear fit of log(pi_a0) ~ log(q): [b0, b1]={}, cov={}".format(
                    popt, pcov
                )
            )
            # TODO: Check if b1 differs a lot: if so, retrain with the same b1
    log_a0_est = torch.as_tensor(linear(n.log().cpu().numpy(), *popt))
    if shrink:
        y = torch.as_tensor(a0.log())
        y[torch.isnan(y)] = log_a0_est[torch.isnan(y)]
        log_a0_est = shrink_normal_normal(
            y, log_a0_est, shrink_prior_var=shrink_prior_var
        )
    return (torch.exp(log_a0_est), popt)


def get_pred_alpha0(X, sample_size_factors, popt, sample_mask=None):
    n_reps, n_condits, n_guides, n_alleles = X.shape
    if sample_mask is None:
        sample_mask = torch.ones((n_reps, n_condits), device="cpu")
    q = get_q(X + 1, sample_size_factors, sample_mask=sample_mask)
    a0_est = np.exp(linear(q.sum(axis=-1).log().cpu().numpy(), *popt))
    return a0_est



import argparse
import gzip

import pandas as pd
from Bio import SeqIO


class InputFileError(Exception):
    """Raised when the input file is not valid."""

    pass


def _check_file(filename):
    """Check if file exists and is readable."""
    try:
        with open(filename, "r"):  # pylint: disable=unspecified-encoding
            pass
    except IOError as exc:
        raise InputFileError(f"I cannot open the file: {filename}") from exc


def _check_library(library_name):
    """Check if library exists."""
    try:
        return __import__(library_name)
    except ImportError as exc:
        raise ImportError(
            f"You need to install {library_name} module to use CRISPRessoCount!"
        ) from exc


def _get_input_parser():
    """Get the input data"""
    print(
        r"""
    _ _       
  /  \ '\                       _   
  |   \  \      __ ___ _  _ _ _| |_ 
   \   \  |    / _/ _ \ || | ' \  _|
    `.__|/     \__\___/\_,_|_||_\__|
    """
    )

    parser = argparse.ArgumentParser(
        description="CRISPRessoCount parameters",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "-b",
        "--edited-base",
        type=str,
        required=True,
        help="For base editors, the base that should be ignored when matching the gRNA sequence",
    )
    parser.add_argument(
        "-f",
        "--sgRNA-filename",
        type=str,
        required=True,
        help="""sgRNA description file. The format requires three columns: name, sequence, barcode [ reporter [,strand, target_pos], [start_pos, offset] ].""",
    )
    # optional
    parser.add_argument(
        "--guide-start-seq",
        type=str,
        help="Guide starts after this sequence in R1",
        default="",
    )
    parser.add_argument(
        "--guide-end-seq",
        type=str,
        help="Guide starts after this sequence in R1",
        default="",
    )
    parser.add_argument(
        "-r", "--count-reporter", help="Count reporter edits.", action="store_true"
    )
    parser.add_argument(
        "-q",
        "--min_average_read_quality",
        type=int,
        help="Minimum average quality score (phred33) to keep a read",
        default=30,
    )
    parser.add_argument(
        "-s",
        "--min-single-bp-quality",
        type=int,
        help="Minimum single bp score (phred33) to keep a read",
        default=0,
    )
    parser.add_argument("-n", "--name", help="Output name", default="")
    parser.add_argument("-o", "--output-folder", help="", default="")
    parser.add_argument(
        "-l", "--reporter-length", type=int, help="length of the reporter", default=32
    )
    parser.add_argument(
        "--keep-intermediate",
        help="Keep all the  intermediate files",
        action="store_true",
    )
    parser.add_argument(
        "--qstart-R1",
        help="Start position of the read when filtering for quality score of the read 1",
        type=int,
        default=0,
    )
    parser.add_argument(
        "--qend-R1",
        help="End position of the read when filtering for quality score of the read 1",
        type=int,
        default=47,
    )
    parser.add_argument(
        "--qstart-R2", help="Same as qstart_R1, for read 2 fastq file", default=0
    )
    parser.add_argument(
        "--qend-R2", help="Same as qstart_R2, for read 2 fastq file", default=36
    )
    parser.add_argument(
        "--gstart-reporter",
        help="Start position of the guide sequence in the reporter",
        type=int,
        default=6,
    )
    parser.add_argument(
        "--match-target-pos",
        help="Count the edit in the exact target position.",
        action="store_true",
    )
    parser.add_argument(
        "--target-pos-col",
        help="Column name specifying the relative target position within *reporter* sequence.",  # ??
        default="target_pos",
    )

    parser.add_argument("--guide-bc", help="Construct has guide barcode", default=True)
    parser.add_argument(
        "--guide-bc-len",
        help="Guide barcode sequence length at the beginning of the R2",
        type=str,
        default=4,
    )
    parser.add_argument(
        "--offset",
        help="Guide file has `offest` column that will be added to the relative position of reporters.",
        action="store_true",
    )
    parser.add_argument(
        "--align-fasta",
        help="gRNA is aligned to this sequence to infer the offset. Can be used when the exact offset is not provided.",
        type=str,
        default="",
    )

    parser.add_argument(
        "--string-allele",
        help="Store allele as quality filtered string instead of Allele object",
        action="store_true",
    )
    parser.add_argument(
        "-g",
        "--count-guide-edits",
        help="count the self editing of guides",
        action="store_true",
    )
    parser.add_argument(
        "-m",
        "--count-guide-reporter-alleles",
        help="count the matched allele of guide and reporter edit",
        action="store_true",
    )
    parser.add_argument(
        "--tiling",
        help="Specify that the guide library is tiling library without 'n guides per target' design",
        action="store_true",
    )

    return parser


def _get_first_read_length(fastq_filename: str) -> int:
    """Obtain length of first read entry of fastq file"""
    if fastq_filename.split(".")[-1] == "gz":
        handle = gzip.open(fastq_filename, "rt")
    else:
        handle = fastq_filename

    for record in SeqIO.parse(handle, "fastq"):
        return len(record)
    raise InputFileError("Provided R1 file doesn't have any read to parse")


def _check_read_length(args: argparse.Namespace, read_length: int, warn_logger):
    """Checks if quality filter positions are compatible with provided read length."""
    if args.qstart_R1 >= read_length or args.qstart_R2 >= read_length:
        raise ValueError(
            f"The start position of base quality filter is not nonnegative ({args.qstart_R1} for R1, {args.qstart_R2} for R2 provided)"
        )
    if args.qend_R1 >= read_length or args.qend_R2 >= read_length:
        raise ValueError(
            f"The start position of base quality filter is not nonnegative ({args.qstart_R1} for R1, {args.qstart_R2} for R2 provided)"
        )
    if args.qend_R2 != args.guide_bc_len + args.reporter_length:
        warn_logger(
            f"Quality of R2 checked up until {args.qend_R2}bp, while the length of guide barcode and reporter combined is {args.guide_bc_len + args.reporter_length}bp."
        )


def _check_arguments(args, info_logger, warn_logger, error_logger):
    """Check the argument validity of the ArgumentParser"""

    if args.sgRNA_filename:
        _check_file(args.sgRNA_filename)

    # Edited base should be one of A/C/T/G
    if args.edited_base.upper() not in ["A", "C", "T", "G"]:
        raise ValueError(
            f"The edited base should be one of A/C/T/G, {args.edited_base} provided."
        )

    edited_base = args.edited_base.upper()
    info_logger(f"Using specified edited base: {edited_base}")
    info_logger(
        f"Using guide barcode length {args.guide_bc_len}, guide start '{args.guide_start_seq}'"
    )
    # normalize name and remove not allowed characters
    # if args.name:
    #     clean_name = slugify(args.name)
    #     if args.name != clean_name:
    #         warn_logger(
    #             f"The specified name {args.name} contained characters not allowed and was changed to: {clean_name}"
    #         )
    #         args.name = clean_name
    sgRNA_info_tbl = pd.read_csv(args.sgRNA_filename)

    def _check_sgrna_info_table(args, sgRNA_info_tbl):
        if args.offset:
            if "offset" not in sgRNA_info_tbl.columns:
                raise InputFileError(
                    "Offset option is set but the input file doesn't contain the `offset` column."
                )
            if len(args.align_fasta) > 0:
                error_logger("Can't have --offset and --align_fasta option together.")
        if (
            args.match_target_pos
            and args.target_pos_col is not None
            and args.target_pos_col not in sgRNA_info_tbl.columns
        ):
            raise InputFileError(
                f"Specified target position column '{args.target_pos_col}' not in the input file {args.sgRNA_filename}."
            )

    _check_sgrna_info_table(args, sgRNA_info_tbl)

    if args.match_target_pos and (args.target_pos_col not in sgRNA_info_tbl.columns):
        raise InputFileError(
            "Target position option is set but the input file doesn't contain the target position column."
        )

    info_logger("Done checking input arguments.")

    return args



# uncompyle6 version 3.8.1.dev0
# Python bytecode 3.8.0 (3413)
# Decompiled from: Python 3.8.11 (default, Aug  3 2021, 15:09:35)
# [GCC 7.5.0]
# Embedded file name: /data/pinello/PROJECTS/2021_08_ANBE/software/bean/bean/framework/ReporterScreen.py
# Compiled at: 2022-01-06 10:00:30
pass
# okay decompiling ReporterScreen.cpython-38.pyc



import numpy as np

def jaccard(set1, set2):
    intersection = len(set(set1).intersection(set2))
    union = (len(set1) + len(set2)) - intersection
    if union == 0: return 1
    return float(intersection) / union


# functionalities for manipulating allele counts
from typing import Dict
from .ReporterScreen import ReporterScreen
from .Edit import Edit
import pandas as pd

def get_edit_from_allele(allele_count_df: pd.DataFrame, allele_col: str = 'allele'):
    """
    Get edit counts from allele counts pd.DataFrame
    -----
    Arguments
    allele_count_df -- DataFrame with guide, allele, and sample columns
    allele_col -- column name of allele_count_df that specifies allele.
    """
    allele_count_df = self.uns[allele_count_key].copy()
    allele_count_df["edits"] = allele_count_df[allele_col].map(lambda a: str(a).split(","))
    allele_count_df = allele_count_df.explode("edits").groupby(["guide", "edits"]).sum()
    allele_count_df = allele_count_df.reset_index().rename(columns={"edits":"edit"})
    def str_to_edit(s):
        try:
            return(Edit.from_str(s))
        except:
            return(AminoAcidEdit.from_str(s))
        # except Exception as e:
        #     print(f"Can't make edit {s} as Edit or AminoAcidEdit")
        #     print(e)
        #     return(s)
    allele_count_df["edit"] = allele_count_df.edit.map(str_to_edit(s))
    return(df)

def simulate_allele_counts(bdata: ReporterScreen, allele_counts_df: pd.DataFrame, allele_col: str = 'allele'):
    edit_counts = get_edit_from_allele(allele_counts_df, allele_col)
    nt_edit_rates = bdata._get_allele_norm(edit_counts_df)
    
    # For each guide, calculate allele proportion. 



# def get_nt_edit_rates(bdata, edit_counts_df: pd.DataFrame, norm_layer_key = "X_bcmatch") -> Dict[str, Edit]:
#     """
#     Get nucleotide level editing rates per guide
#     """
#     nt_edit_rates = bdata._get_allele_norm(edit_counts_df)
#     nt_edit_rates.
    


import pickle as pkl
import pandas as pd
import beret as be
from .post_training import get_result_table

def get_target_info(path):
    sim = be.read_h5ad(path)
    guide_info = sim.guides.copy()
    edit_rate_info = guide_info[['target', "effect_size", 'edit_rate', 'coverage']].groupby(
            'target', sort=False).agg({"edit_rate":["mean", "std"], 'coverage':'mean'})
    edit_rate_info.columns = ["edit_rate_mean", "edit_rate_std", "coverage_mean"]
    target_info = guide_info[['target', "effect_size"]].drop_duplicates().set_index("target", drop=True)
    target_info = target_info.join(edit_rate_info).reset_index()
    return(guide_info, target_info)

def get_multiple_result_table(adata_path, prefix, model_ids, path_pattern_string = "{}.model{}.result.pkl"):
    guide_info, target_info = get_target_info(adata_path)

    model_res = {}

    for model_id in model_ids:
        f = open(path_pattern_string.format(prefix, model_id), "rb")
        model_res[model_id] = pkl.load(f)
        f.close()

    var_dfs = {}
    sgrna_dfs = {}
    for model_id, d in model_res.items():
        res = get_result_table(target_info, d, guide_info.index, f"_{model_id}")
        var_dfs[model_id], sgrna_dfs[model_id] = res

    df = pd.concat(var_dfs.values(), axis=1)
    res_df = df.loc[:,~df.columns.duplicated()].copy()
    return res_df

def read_mageck_result(prefix, models = ['topbot', 'topbot_var', 'sort', 'sort_var'], suffix=""):
    tbls = []
    for model in models:
        df = pd.read_csv("{}/{}.gene_summary.txt".format(prefix, model), sep = "\t")
        df.columns = [s.split("|")[-1] for s in df.columns]
        df = df.add_suffix(f"_{model}{suffix}")
        tbls.append(df)
    return(pd.concat(tbls, axis=1))
    


from typing import Sequence, Union, Optional, Dict
import numpy as np
import torch
import pandas as pd
from scipy.stats import norm
from statsmodels.stats.multitest import fdrcorrection

def get_sd_quantile(quantiles):
    # 16-84% quantile is 2 sigma.
    return (quantiles[83]-quantiles[15])/2

def get_fdr(mu_z: Union[np.ndarray, torch.Tensor], plot: bool=False, get_p = False):
    """Return B-H corrected p-values of normal distribution based on z-score
    
    Args
        mu_z: Tensor of z-scores
        plot: If True, print gnu histogram plot
    """
    if isinstance(mu_z, torch.Tensor): mu_z = mu_z.detach().numpy()
    p_dec = norm.cdf(mu_z)
    p_inc = 1-norm.cdf(mu_z)
    # if plot:
    #     gp.plot(p_dec, dict(histogram = 'freq'))
    #     gp.plot(p_inc, dict(histogram = 'freq'))
    _, fdr_dec = fdrcorrection(p_dec)
    _, fdr_inc = fdrcorrection(p_inc)
    fdr = np.minimum(fdr_dec, fdr_inc)
    if not get_p:
        return(fdr_dec, fdr_inc, fdr)
    else:
        return(fdr_dec, fdr_inc, fdr, p_dec, p_inc, np.minimum(p_dec, p_inc))

def write_result_table(
    target_info_df: pd.DataFrame, 
    param_hist_dict: Dict[str,Union[Dict,torch.Tensor]], 
    prefix: str="", 
    write_fitted_eff: Optional[bool] = True, 
    guide_index: Optional[Sequence]=None, 
    guide_acc: Optional[Sequence[float]] = None, 
    return_result: Optional[bool]=False,
    ) -> Union[None, pd.DataFrame]:
    """Write fitted result into text files.
    """
    try:
        mu = param_hist_dict['quantiles']['mu_alleles'][50].detach()[:,0].numpy()
        mu_sd = get_sd_quantile(param_hist_dict['quantiles']['mu_alleles'].detach())[:,0].numpy()
        sd = param_hist_dict['quantiles']['sd_alleles'][50].detach()[:,0].numpy()
    except:
        mu = param_hist_dict['params']['mu_loc'].detach()[:,0].numpy()
        mu_sd = param_hist_dict['params']['mu_scale'].detach()[:,0].numpy()
        sd = param_hist_dict['params']['sd_loc'].detach().exp()[:,0].numpy()
    mu_z = mu/mu_sd 
    fdr_dec, fdr_inc, fdr = get_fdr(mu_z)
    fit_df = pd.DataFrame({
        "mu": mu,
        "mu_sd": mu_sd,
        "mu_z": mu/mu_sd,
        "sd": sd,
        "fdr_dec": fdr_dec,
        "fdr_inc":fdr_inc,
        "fdr":fdr
        }
    )
    if "negctrl" in param_hist_dict.keys():
        print("Normalizing with common negative control distribution")
        try:
            mu0 = param_hist_dict['negctrl']['params']['mu_loc'].detach().numpy()
            sd0 = param_hist_dict['negctrl']['params']['sd_loc'].detach().numpy()
        except KeyError:
            print(param_hist_dict['negctrl']['params'].keys())
        print(f"Fitted mu0={mu0}, sd0={sd0}.")
        fit_df['mu_adj'] = (mu - mu0)/sd0
        fit_df['mu_sd_adj'] = mu_sd / sd0
        fit_df['mu_z_adj'] = fit_df.mu_adj / fit_df.mu_sd_adj
        fit_df['sd_adj'] = sd / sd0
        fdr_dec, fdr_inc, fdr = get_fdr(fit_df.mu_z_adj)
        fit_df['fdr_dec_adj'], fit_df['fdr_inc_adj'], fit_df['fdr_adj'] = fdr_dec, fdr_inc, fdr
    fit_df = pd.concat([target_info_df, fit_df], axis=1)
    if return_result: return(fit_df)
    fit_df.to_csv(f"{prefix}CRISPRbean_element_result.csv")

    if write_fitted_eff or not guide_acc is None:
        if not "alpha_pi" in param_hist_dict['params'].keys():
            pi=1.0
        else:
            a_fitted = param_hist_dict['params']['alpha_pi'].detach().numpy()
            pi = a_fitted[...,0]/a_fitted.sum(axis=1)
        try:
            sgRNA_df = pd.DataFrame({"edit_eff":pi}, index=guide_index)
        except ValueError:
            print(pi.shape)
        if not guide_acc is None:
            sgRNA_df['accessibility'] = guide_acc
        sgRNA_df.to_csv(f"{prefix}CRISPRbean_sgRNA_result.csv")



from typing import Dict
import torch
import numpy as np
import pyBigWig
import pandas as pd
import anndata as ad
import bean as be


class Alias:
    def __init__(self, source_name):
        self.source_name = source_name

    def __get__(self, obj, objtype=None):
        if obj is None:
            # Class lookup, return descriptor
            return self
        return getattr(obj, self.source_name)

    def __set__(self, obj, value):
        setattr(obj, self.source_name, value)


def _get_accessibility_single(
    pos: int,
    track,
    chrom: str = "chr19",
    guide_start_pos=0,
    half_window_size: int = 100,
):
    """Obtain mean log accessibility signal for a given genomic position, padded by `half_window_size`.

    Args
    --
    pos: genomic position
    track: pyBigWig.bigWigFile object with accessibility signal information
    chrom: chromosome name in the bigWig file.
    guide_start_pos: offset to the genomic position to be added
    half_window_size: half-size of the window centered at pos + guide_start_pos of which the signal would be taken
    """
    if half_window_size < 0:
        raise ValueError("Window size must be non-negative.")
    if pos == "control" or np.isnan(pos):
        return np.nan
    try:
        return np.exp(
            np.nanmean(
                np.log(
                    np.array(
                        track.values(
                            chrom,
                            int(guide_start_pos + pos - half_window_size),
                            int(guide_start_pos + pos + half_window_size),
                        )
                    )
                    + 1.0
                )
            )
        )
    except Exception as e:
        print(e)
        return np.nan


def get_accessibility_guides(
    accessibility_bw_path: str, guide_info: pd.DataFrame, half_window_size: int = 100
) -> torch.Tensor:
    """Gets guide accessibility given a dataframe with `genomic_pos`, `chr` column. Returns the mean logged signal of the guide position padded by `half_window_size`.

    Args:
    accessibility_bw_path: bigWig file path with ATAC-seq signal.
    guide_info: screen.guides dataframe with `genomic_pos`, `chr` column.
    half_window_size: For each guide, accessibility signal its position padded by this half_window_size upstream and downstream
    """
    acc = pyBigWig.open(accessibility_bw_path)
    guide_accessibility = torch.as_tensor(
        guide_info.apply(
            lambda row: _get_accessibility_single(
                row.genomic_pos,
                acc,
                chrom=row.chr,
                guide_start_pos=0,
                half_window_size=half_window_size,
            ),
            axis=1,
        )
    )
    guide_accessibility[torch.isnan(guide_accessibility)] = torch.nanmedian(
        guide_accessibility
    )
    return guide_accessibility


def get_edit_to_index_dict(cnalleles: pd.Series) -> Dict[str, int]:
    """Returns mapping (dictionary) of edit to unique index
    Returns Dict[edit(str) -> index(int)] for edits
    ---
    Arguments
    cnalleles -- pd.Series object of CodingNoncodingAllele objects.
    """
    edit_lists = cnalleles.map(
        lambda a: list(a.aa_allele.edits) + list(a.nt_allele.edits)
    )
    edits = pd.Series(
        pd.Series(
            [e.get_abs_edit() for l in edit_lists.tolist() for e in l], dtype="object"
        ).unique()
    )
    edits_to_index = pd.Series(edits)
    edits_to_index = (
        edits_to_index.reset_index()
        .rename(columns={0: "edit"})
        .set_index("edit")["index"]
        .to_dict()
    )
    return edits_to_index


def _insert_row_(row_number: int, df: pd.DataFrame, row):
    # https://www.geeksforgeeks.org/insert-row-at-given-position-in-pandas-dataframe/
    # Function to insert row in the dataframe
    df1 = df.iloc[0:row_number]
    df2 = df.iloc[row_number:]
    return pd.concat([df1, row, df2], axis=0)


def _insert_row_to_obs(adata: be.ReporterScreen, row_number: int, row_value, index):
    obs = _insert_row_(row_number, adata.guides, row_value)
    X = np.insert(adata.X, row_number, 0, axis=0)
    layers = {}
    for layer in adata.layers.keys():
        layers[layer] = np.insert(adata.layers[layer], row_number, 0, axis=0)
    if "repguide_mask" in adata.uns:
        adata.uns["repguide_mask"] = pd.concat(
            [
                adata.uns["repguide_mask"].iloc[:row_number, :],
                pd.DataFrame(
                    columns=adata.uns["repguide_mask"].columns,
                    index=[index],
                ).fillna(0),
                adata.uns["repguide_mask"].iloc[row_number:, :],
            ],
            axis=0,
        )
    adata = be.ReporterScreen(
        X=X,
        guides=obs,
        samples=adata.samples,
        uns=adata.uns,
        obsm=adata.obsm,
        varm=adata.varm,
        layers=layers,
    )
    return adata


def check_consecutive_targets(target_list, guide_per_target_counts=5):
    item_count = 0
    prev_item = ""
    for item in target_list:
        if item_count == 0:
            item_count += 1
            prev_item = item
        else:
            if item == prev_item:
                item_count += 1
            else:
                assert (
                    item_count == guide_per_target_counts
                ), f"not all targets are consecutive in len {guide_per_target_counts}"
                item_count = 1
                prev_item = item


def _assign_rep_ids_and_sort(
    screen: be.ReporterScreen, rep_col: str, condition_column: str = None
) -> be.ReporterScreen:
    """Assign replicate IDs to samples and sort them accordingly."""
    for i, rep in enumerate(sorted(screen.samples[rep_col].unique())):
        screen.samples.loc[screen.samples[rep_col] == rep, f"{rep_col}_id"] = i
        if condition_column is None:
            sort_key = f"{rep_col}_id"
        else:
            sort_key = [f"{rep_col}_id", f"{condition_column}_id"]
        screen = screen[
            :,
            screen.samples.sort_values(sort_key).index,
        ]
    return screen


def _obtain_effective_edit_rate(ndata, count_thres=10):
    """Calculates effective editing rate of each variant for tiling screen.

    Allele count
    Args
    --
    count_thres: Per-replicate allele editing rate is ignored if X_bcmatch is lower than this threshold.
    """
    allele_rates = ndata.allele_counts_control / ndata.X_bcmatch_control[:, :, :, None]
    allele_rates[
        (ndata.X_bcmatch_control < count_thres)[:, :, :, None].expand(
            allele_rates.shape
        )
    ] = torch.nan
    mean_allele_rates = allele_rates.nanmean(axis=(0, 1))[:, 1:]
    assert mean_allele_rates.shape == (ndata.n_guides, ndata.n_max_alleles - 1)
    allele_to_edit_normed = (
        ndata.allele_to_edit / ndata.allele_to_edit.nansum(axis=-1)[:, :, None]
    )

    allele_to_edit_normed.shape == (
        ndata.n_guides,
        ndata.n_max_alleles - 1,
        ndata.n_edits,
    )
    return (mean_allele_rates[:, :, None] * allele_to_edit_normed).nansum(axis=(0, 1))


def _obtain_n_guides_alleles_per_variant(ndata):
    return (ndata.allele_to_edit.sum(axis=1) > 0).sum(axis=0)



from turtle import position
from typing import List, Union
import subprocess as sb
import numpy as np
import pandas as pd
import gzip
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqIO.QualityIO import FastqGeneralIterator
from .CRISPResso2Align import read_matrix, global_align_base_editor
from ..framework.Edit import Allele, Edit


class InputFileError(Exception):
    pass


def _base_edit_to_from(start_base: chr = "A"):
    try:
        base_map = {"A": "G", "C": "T"}
    except KeyError:
        raise ValueError("Only A/C are supported for base to be edited.")
    return base_map[start_base]


def _read_count_match(R1_filename: str, R2_filename: str) -> int:
    R1_count = _get_n_reads_fastq(R1_filename)
    R2_count = _get_n_reads_fastq(R2_filename)
    if R1_count != R2_count:
        raise InputFileError(
            "Paired end read numbers are different in R1({}) and R2({})".format(
                R1_count, R2_count
            )
        )
    return R1_count


def _get_n_reads_fastq(fastq_filename: str):
    p = sb.Popen(
        ("z" if fastq_filename.endswith(".gz") else "")
        + "cat < %s | wc -l" % fastq_filename,
        shell=True,
        stdout=sb.PIPE,
    )
    return int(float(p.communicate()[0]) / 4.0)


def _get_fastq_handle(fastq_filename: str, mode="r"):
    if fastq_filename.endswith(".gz"):
        if mode == "r":
            mode = "rt"
        elif mode == "w":
            mode = "wb"
        fastq_handle = gzip.open(fastq_filename, mode)
    else:
        fastq_handle = open(fastq_filename, mode)
    return fastq_handle


def _read_is_good_quality(
    record: SeqIO.SeqRecord, min_bp_quality=0, min_single_bp_quality=0, qend=-1
):
    mean_quality_pass = (
        np.array(record.letter_annotations["phred_quality"])[:qend].mean()
        >= min_bp_quality
    )
    min_quality_pass = (
        np.array(record.letter_annotations["phred_quality"])[:qend].min()
        >= min_single_bp_quality
    )
    return mean_quality_pass and min_quality_pass


def _check_readname_match(R1: List[SeqIO.SeqRecord], R2: List[SeqIO.SeqRecord]):
    if len(R1) != len(R2):
        raise ValueError("The number of reads in R1 and R2 file does not match.")

        for i in range(len(R1)):
            R1_record = R1[i]
            R2_record = R2[i]

            if R1_record.name != R2_record.name:
                raise InputFileError(
                    "R1 and R2 read discordance in read {} and {}".format(
                        R1_record.name, R2_record.name
                    )
                )


def _get_guide_to_reporter_df(sgRNA_filename: str) -> pd.DataFrame:
    """Returns a gRNA name to reporter sequence mapping."""
    guide_to_reporter = {}

    with open(sgRNA_filename) as infile:
        sgRNA_df = pd.read_csv(infile)
        if not ("name" in sgRNA_df.columns and "Reporter" in sgRNA_df.columns):
            raise InputFileError(
                "Input gRNA file doesn't have the column 'gRNA' or 'gRNA_barcode'."
            )
        sgRNA_df.set_index("name", inplace=True)
        return sgRNA_df


def revcomp(seq: Union[Seq, str]) -> str:
    if isinstance(seq, str):
        seq = Seq(seq)
    return str(seq.reverse_complement())


def _fastq_iter_to_text(record: SeqIO.SeqRecord):
    t = (record.id,)
    seq = record.seq
    q = record.letter_annotations["phred_quality"]
    return "{}\n{}\n+\n{}\n".format(t, seq, q)


def _write_paired_end_reads(R1_record, R2_record, R1_out_handle, R2_out_handle):
    try:
        R1_out_handle.write(_fastq_iter_to_text(R1_record))
        R2_out_handle.write(_fastq_iter_to_text(R2_record))
    except TypeError:
        R1_out_handle.write(_fastq_iter_to_text(R1_record).encode())
        R2_out_handle.write(_fastq_iter_to_text(R2_record).encode())


def _get_edited_allele(
    ref_seq: str,
    query_seq: str,
    offset: int,
    strand: int = 1,
    start_pos: int = 0,
    end_pos: int = 100,
):

    allele = Allele()

    assert len(ref_seq) == len(query_seq), "reference and query seq length mismatch"

    for i, (ref_nt, sample_nt) in enumerate(zip(ref_seq, query_seq)):
        if i < start_pos or i >= end_pos:
            continue
        if ref_nt == sample_nt:
            continue
        else:
            edit = Edit(i - start_pos, ref_nt, sample_nt, offset, strand=strand)
            allele.add(edit)

    return allele


def _write_alignment_matrix(
    ref_base: str, alt_base: str, path, allow_complementary=False
):
    """
    Writes base substitution matrix
    """
    bases = ["A", "C", "T", "G"]
    if not (ref_base in bases and alt_base in bases):
        raise ValueError(
            "Specified ref base '{}' or alt base '{}' isn't valid".format(
                ref_base, alt_base
            )
        )
    mat = np.ones((4, 4), dtype=int) * -4
    np.fill_diagonal(mat, 5)
    aln_df = pd.DataFrame(mat, index=bases, columns=bases)
    aln_df.loc[ref_base, alt_base] = 0
    if allow_complementary:
        comp_map = {"A": "T", "C": "G", "T": "A", "G": "C"}
        aln_df.loc[comp_map[ref_base], comp_map[alt_base]] = 0
    aln_df.to_csv(path, sep=" ")


def _get_allele_from_alignment(
    ref_aligned: str,
    query_aligned: str,
    offset: int,
    strand: int,
    start_pos: int,
    end_pos: int,
    positionwise_quality: np.ndarray = None,
    quality_thres: float = -1,
):
    assert len(ref_aligned) == len(query_aligned)
    allele = Allele()
    ref_gaps = 0
    alt_gaps = 0
    alt_seq_len = len(query_aligned) - query_aligned.count("-")
    if positionwise_quality is None:
        # alt_position_is_good_quality = np.ones(alt_seq_len, dtype=bool)
        alt_position_is_good_quality = np.array(
            [c != "N" for c in query_aligned.replace("-", "")]
        )
    else:
        assert len(positionwise_quality) == alt_seq_len
        alt_position_is_good_quality = positionwise_quality > quality_thres
    for i in range(len(ref_aligned)):
        if ref_aligned[i] == query_aligned[i]:
            continue
        ref_base = ref_aligned[i]
        alt_base = query_aligned[i]
        if alt_base != "-":
            alt_base_is_good_quality = alt_position_is_good_quality[i - alt_gaps]
        else:
            alt_base_is_good_quality = True
        if ref_base == "-":
            ref_gaps += 1
        elif alt_base == "-":
            alt_gaps += 1
        ref_pos = i - ref_gaps
        if alt_base_is_good_quality and ref_pos >= start_pos and ref_pos < end_pos:
            allele.add(
                Edit(
                    rel_pos=ref_pos,
                    ref_base=ref_base,
                    alt_base=alt_base,
                    offset=offset,
                    strand=strand,
                )
            )
    return allele


def _get_edited_allele_crispresso(
    ref_seq: str,
    query_seq: str,
    ref_base: str,
    alt_base: str,
    aln_mat_path: str,
    offset: int,
    strand: int = 1,
    start_pos: int = 0,
    end_pos: int = 100,
    positionwise_quality: np.ndarray = None,
    quality_thres: float = 30,
    objectify_allele=True,
):
    aln_matrix = read_matrix(aln_mat_path)
    assert strand in [-1, +1]
    gap_incentive = np.zeros(len(ref_seq) + 1, dtype=np.int)
    query_aligned, ref_aligned, score = global_align_base_editor(
        query_seq,
        ref_seq,
        ref_base,
        alt_base,
        aln_matrix,
        gap_incentive,
        gap_open=-20,
        gap_extend=-10,
    )
    if objectify_allele:
        allele = _get_allele_from_alignment(
            ref_aligned,
            query_aligned,
            offset,
            strand,
            start_pos,
            end_pos,
            positionwise_quality,
            quality_thres,
        )
        for e in allele.edits:
            if e.ref_base == "-":
                continue
            assert ref_seq[e.rel_pos] == e.ref_base, (
                "relative position mismatch: ref pos {}: {} vs {}".format(
                    e.rel_pos, ref_seq[e.rel_pos], e.ref_base
                )
                + "\nallele {}, \nrefseq {}, \naltseq {}, \n".format(
                    allele, ref_seq, query_seq
                )
                + "ref_align {}, \nalt_align {}".format(ref_aligned, query_aligned)
            )
    else:
        allele = _string_filter_basewise_quality(
            ref_aligned, query_aligned, positionwise_quality, quality_thres
        )
    return (allele, score)


def _string_filter_basewise_quality(
    ref_seq, query_seq, positionwise_quality, quality_thres
):
    query_seq_chars = list(query_seq)
    for i in range(len(positionwise_quality)):
        if positionwise_quality[i] < quality_thres:
            query_seq_chars[i] = ref_seq[i]
    return "".join(query_seq_chars)


def _multiindex_dict_to_df(input_dict, key_column_names, value_column_name):
    if not isinstance(key_column_names, list):
        key_column_names = [key_column_names]
    mi = pd.MultiIndex.from_tuples(
        input_dict.keys(), names=["guide"] + key_column_names
    )
    df = pd.DataFrame.from_dict(
        input_dict,
        orient="index",
        columns=[value_column_name],
    )
    df.index = mi
    df.reset_index(inplace=True)
    if len(key_column_names) == 1:
        df.rename(
            columns={"level_0": "guide", "level_1": key_column_names[0]}, inplace=True
        )
    elif len(key_column_names) == 2:
        df.rename(
            columns={
                "level_0": "guide",
                "level_1": key_column_names[0],
                "level_2": key_column_names[1],
            },
            inplace=True,
        )
    return df



from typing import Dict, Sequence
import pandas as pd
from .utils import get_fdr, get_sd_quantile

def get_result_table(target_info_df: pd.DataFrame, param_hist_dict: Dict, guide_index: Sequence=None, suffix_fitted:str = ""):
    """ Return fitted result as table

    Args
        suffix_fitted: suffix the fitted parameters
    """
    try:
        mu = param_hist_dict['quantiles']['mu_alleles'][50].detach()[:,0].numpy()
        mu_sd = get_sd_quantile(param_hist_dict['quantiles']['mu_alleles'].detach())[:,0].numpy()
        sd = param_hist_dict['quantiles']['sd_alleles'][50].detach()[:,0].numpy()
    except:
        mu = param_hist_dict['params']['mu_loc'].detach()[:,0].numpy()
        mu_sd = param_hist_dict['params']['mu_scale'].detach()[:,0].numpy()
        sd = param_hist_dict['params']['sd_loc'].detach().exp()[:,0].numpy()
    mu_z = mu/mu_sd
    fdr_dec, fdr_inc, fdr, p_dec, p_inc, p = get_fdr(mu_z, get_p = True)
    fit_df = pd.DataFrame({
        f"mu{suffix_fitted}": mu,
        f"mu_sd{suffix_fitted}": mu_sd,
        f"mu_z{suffix_fitted}": mu/mu_sd,
        f"sd{suffix_fitted}": sd,
        f"p_dec{suffix_fitted}": p_dec,
        f"p_inc{suffix_fitted}":p_inc,
        f"p{suffix_fitted}":p,
        f"fdr_dec{suffix_fitted}": fdr_dec,
        f"fdr_inc{suffix_fitted}":fdr_inc,
        f"fdr{suffix_fitted}":fdr
        }
    )
    if "negctrl" in param_hist_dict.keys():
        print("Normalizing with common negative control distribution")
        mu0 = param_hist_dict['negctrl']['mu_loc']
        sd0 = param_hist_dict['negctrl']['sd_loc']
        print(f"Fitted mu0={mu0:3f}, sd0={sd0:3f}.")
        fit_df[f'mu_adj{suffix_fitted}'] = (mu - mu0)/sd0
        fit_df[f'mu_sd_adj{suffix_fitted}'] = mu_sd / sd0
        fit_df[f'sd_adj{suffix_fitted}'] = sd / sd0
    fit_df = pd.concat([target_info_df, fit_df], axis=1)

    if not "alpha_pi" in param_hist_dict['params'].keys():
        print("Model didn't have fitted editing rate. Not writing the sgRNA result file.")
        sgRNA_df = None
    else:
        a_fitted = param_hist_dict['params']['alpha_pi'].detach().numpy()
        pi = a_fitted[:,0]/a_fitted.sum(axis=1)
        sgRNA_df = pd.DataFrame({f"edit_eff{suffix_fitted}":pi}, index=guide_index)
    return(fit_df, sgRNA_df)

def write_result_table(target_info_df, param_hist_dict, prefix="", guide_index=None, suffix_fitted = ""):
    """ Write fitted result as table

    Args
        suffix_fitted: suffix the fitted parameters in the resulting table
    """
    fit_df, sgRNA_df = get_result_table(target_info_df, param_hist_dict, prefix=prefix, guide_index=guide_index, suffix_fitted=suffix_fitted)
    fit_df.to_csv(f"{prefix}_variant_result.csv")
    if sgRNA_df: sgRNA_df.to_csv(f"{prefix}_CRISPRbean_sgRNA_result.csv")
    else:
        print("No sgRNA info fitted.")



from copy import deepcopy
from typing import Collection, Iterable, List, Optional, Union, Sequence, Literal
import re
import anndata as ad
import numpy as np
import pandas as pd
from perturb_tools import Screen

from ..annotate._supporting_fn import (
    filter_allele_by_base,
    filter_allele_by_pos,
    get_aa_alleles,
)
from .AminoAcidEdit import AminoAcidEdit, CodingNoncodingAllele
from .Edit import Allele, Edit
from ..annotate.filter_alleles import (
    _distribute_alleles_to_filtered,
    _map_alleles_to_filtered,
)


def _get_counts(
    filename_pattern: str,
    guides: pd.DataFrame,
    reps: Iterable[str],
    samples: Iterable[str],
) -> pd.DataFrame:
    """Read multiple .csv file with "guide_name,read_counts" columns across replicates and samples.
    File names are dictated by filename_pattern and reps, condiitons argument provided.

    Args
    --
    filename_pattern: format string with 'r' as replicate and 'c' as condition. Will be formatted as filename_pattern.format(r=replicate, c=condition).
    guides: Guide information that will be used as reference
    reps: List of replicate string to be used for formatting
    samples: List of condition strings to be used for formatting
    """
    for rep in reps:
        for spl in samples:
            res = pd.read_csv(filename_pattern.format(r=rep, c=spl), delimiter="\t")
            res = res.set_index("guide_name")[["read_counts"]]
            res.columns = res.columns.map(lambda x: "{rep}_{spl}")
            guides = guides.join(res, how="left")
        guides = guides.fillna(0)
        return guides


def _get_edits(
    filename_pattern: str,
    guide_info: pd.DataFrame,
    reps: Iterable[str],
    samples: Iterable[str],
    count_exact: bool = True,
) -> pd.DataFrame:
    """ """
    edits = pd.DataFrame(index=(guide_info.index))
    for rep in reps:
        for spl in samples:
            res = pd.read_csv(filename_pattern.format(r=rep, c=spl), delimiter="\t")
            res = res.set_index("name")
            if count_exact:
                res_info = guide_info.join(res, how="left").reset_index()
                this_edit = res_info.loc[
                    (
                        (res_info["Target base position in gRNA"] == res_info.pos + 1)
                        & (res_info.ref_base == "A")
                        & (res_info.sample_base == "G")
                    )
                ][["name", "count"]].set_index("name", drop=True)["count"]
            else:
                this_edit = (
                    res.loc[(res.ref_base == "A") & (res.sample_base == "G"), :]
                    .groupby("name")["count"]
                    .sum()
                )
            this_edit.name = "{r}_{c}".format(r=rep, c=spl)
            edits = edits.join(this_edit, how="left")
        edits = edits.fillna(0)
        return edits


class ReporterScreen(Screen):
    def __init__(
        self,
        X=None,
        X_edit=None,
        X_bcmatch=None,
        target_base_change: Optional[str] = None,
        replicate_label: str = "rep",
        condition_label: str = "bin",
        tiling: Optional[bool] = None,
        *args,
        **kwargs,
    ):
        (super().__init__)(X=X, *args, **kwargs)
        if X_edit is not None:
            self.layers["edits"] = X_edit
        if X_bcmatch is not None:
            self.layers["X_bcmatch"] = X_bcmatch
        for k, df in self.uns.items():
            if not isinstance(df, pd.DataFrame):
                continue
            if "guide" in df.columns and len(df) > 0:
                if (
                    "allele" in df.columns
                    and (not isinstance(df["allele"].iloc[0], Allele))
                    and any(df["allele"].map(Allele.match_str))
                ):
                    self.uns[k].loc[:, "allele"] = self.uns[k].allele.map(
                        lambda s: Allele.from_str(s)
                    )
                if (
                    "edit" in df.columns
                    and (not isinstance(df["edit"].iloc[0], Edit))
                    and any(df["edit"].map(Edit.match_str))
                ):
                    self.uns[k].loc[:, "edit"] = self.uns[k].edit.map(
                        lambda s: Edit.from_str(s)
                    )
                if "reporter_allele" in df.columns and "guide_allele" in df.columns:
                    self.uns[k].loc[:, "reporter_allele"] = self.uns[
                        k
                    ].reporter_allele.map(lambda s: Allele.from_str(s))
                    self.uns[k].loc[:, "guide_allele"] = self.uns[k].guide_allele.map(
                        lambda s: Allele.from_str(s)
                    )
                if "aa_allele" in df.columns and any(
                    df["aa_allele"].map(CodingNoncodingAllele.match_str)
                ):
                    self.uns[k].loc[:, "aa_allele"] = self.uns[k].aa_allele.map(
                        lambda s: CodingNoncodingAllele.from_str(s)
                    )
        if target_base_change is not None:
            if not re.fullmatch(r"[ACTG]>[ACTG]", target_base_change):
                raise ValueError(
                    f"target_base_change {target_base_change} doesn't match the allowed base change. Feed in valid base change string ex) 'A>G', 'C>T'"
                )
            self.uns["target_base_change"] = target_base_change
        if tiling is not None:
            self.uns["tiling"] = tiling

    @property
    def X_edits(self):
        return self.layers["edits"]

    @property
    def X_bcmatch(self):
        return self.layers["X_bcmatch"]

    @property
    def edit_tables(self):
        return {k: self.uns[k] for k in self.uns.keys() if "edit" in k}

    @property
    def allele_tables(self):
        return {k: self.uns[k] for k in self.uns.keys() if "allele" in k}

    @property
    def base_edited_from(self):
        return self.uns["target_base_change"][0]

    @property
    def base_edited_to(self):
        return self.uns["target_base_change"][-1]

    @property
    def target_base_change(self):
        return self.uns["target_base_change"]

    @property
    def tiling(self):
        return self.uns["tiling"]

    @classmethod
    def from_file_paths(
        cls,
        reps: List[str] = None,
        samples: List[str] = None,
        guide_info_file_name: str = None,
        guide_count_filenames: str = None,
        guide_bcmatched_count_filenames: str = None,
        edit_count_filenames: str = None,
    ):
        guide_info = pd.read_csv(guide_info_file_name).set_index("name")
        guides = pd.DataFrame(index=(pd.read_csv(guide_info_file_name)["name"]))
        guides_lenient = _get_counts(guide_count_filenames, guides, reps, samples)
        edits_ag = _get_edits(
            edit_count_filenames, guide_info, reps, samples, count_exact=False
        )
        edits_exact = _get_edits(edit_count_filenames, guide_info, reps, samples)
        if guide_bcmatched_count_filenames is not None:
            guides_bcmatch = _get_counts(
                guide_bcmatched_count_filenames, guides, reps, samples
            )
            repscreen = cls(
                guides_lenient, edits_exact, X_bcmatch=guides_bcmatch, guides=guide_info
            )
        else:
            repscreen = cls(guides_lenient, edits_exact, guides=guide_info)
        repscreen.layers["edits_ag"] = edits_ag
        repscreen.samples["replicate"] = np.repeat(reps, len(samples))
        repscreen.samples["sort"] = np.tile(samples, len(reps))
        repscreen.uns["replicates"] = reps
        repscreen.uns["samples"] = samples
        return repscreen

    @classmethod
    def from_adata(cls, adata):
        X_bcmatch = adata.layers["X_bcmatch"] if "X_bcmatch" in adata.layers else None
        edits = adata.layers["edits"] if "edits" in adata.layers else None
        return cls(
            (adata.X),
            X_edit=edits,
            X_bcmatch=X_bcmatch,
            guides=(adata.obs),
            samples=(adata.var),
            uns=(adata.uns),
            layers=(adata.layers),
        )

    def copy(self):
        adata = super().copy()
        return type(self).from_adata(adata)

    def rename(self, new_index: Sequence[str], axis: Literal[0, 1], keep_old=False):
        """Rename index"""
        if axis == 0:
            if len(new_index) != len(self.guides):
                raise ValueError(
                    f"New index of length {len(new_index)} doesn't match original index length of {len(self.guides)}."
                )
            index_subs_map = {
                self.guides.index[i]: new_index[i] for i in range(len(new_index))
            }
            if keep_old:
                self.guides["old_index"] = self.guides.index
            self.guides.index = new_index
            for k in self.uns.keys():
                if "guide" in self.uns[k].columns:
                    self.uns[k].guide = self.uns[k].guide.map(index_subs_map)
        else:
            if len(new_index) != len(self.samples):
                raise ValueError(
                    f"New index of length {len(new_index)} doesn't match original index length of {len(self.samples)}."
                )
            index_subs_map = {
                self.samples.index[i]: new_index[i] for i in range(len(new_index))
            }
            if keep_old:
                self.samples["old_index"] = self.samples.index
            self.samples.index = new_index
            for k in self.uns.keys():
                if "guide" in self.uns[k].columns and (
                    "edit" in self.uns[k].columns[1]
                    or "allele" in self.uns[k].columns[1]
                ):
                    self.uns[k].columns = (
                        self.uns[k].columns[:2].tolist()
                        + self.uns[k].columns[2:].map(index_subs_map).tolist()
                    )

    def __add__(self, other):
        if all(self.guides.index == other.guides.index) and all(
            self.samples.index == other.samples.index
        ):
            added_uns = {}
            for k in self.uns.keys():
                if k not in other.uns.keys():
                    continue
                if k == "allele_counts":
                    index_pair = ["guide", "allele"]
                elif k in ["guide_edit_counts", "edit_counts"]:
                    index_pair = ["guide", "edit"]
                elif k == "guide_reporter_allele_counts":
                    index_pair = ["guide", "reporter_allele", "guide_allele"]
                else:
                    continue
                self_df = self.uns[k].set_index(index_pair, drop=True)
                add_df = other.uns[k].set_index(index_pair, drop=True)
                add_df = add_df.loc[:, self_df.columns]
                added_uns[k] = (
                    self_df.add(add_df, fill_value=0).astype(int).reset_index()
                )

            return (
                ReporterScreen(
                    (self.X + other.X),
                    (self.layers["edits"] + other.layers["edits"]),
                    (self.layers["X_bcmatch"] + other.layers["X_bcmatch"]),
                    guides=(self.guides),
                    samples=(self.samples),
                    uns=added_uns,
                )
                if "X_bcmatch" in self.layers and "X_bcmatch" in other.layers
                else ReporterScreen(
                    (self.X + other.X),
                    (self.layers["edits"] + self.layers["edits"]),
                    guides=(self.guides),
                    samples=(self.samples),
                    uns=added_uns,
                )
            )
        raise ValueError("Guides/sample description mismatch")

    def __getitem__(self, index):
        """TODO: currently the samples names are in ['index'] column. Making it to be the index will
        allow the subsetting by condition names.
        """
        oidx, vidx = self._normalize_indices(index)
        if isinstance(oidx, (int, np.integer)):
            if not (-self.n_obs <= oidx < self.n_obs):
                raise IndexError(f"Observation index `{oidx}` is out of range.")
            oidx += self.n_obs * (oidx < 0)
            oidx = slice(oidx, oidx + 1, 1)
        if isinstance(vidx, (int, np.integer)):
            if not (-self.n_vars <= vidx < self.n_vars):
                raise IndexError(f"Variable index `{vidx}` is out of range.")
            vidx += self.n_vars * (vidx < 0)
            vidx = slice(vidx, vidx + 1, 1)
        guides_include = self.guides.iloc[oidx].index.tolist()
        condit_include = self.samples.iloc[vidx].index.tolist()
        adata = super().__getitem__(index)
        new_uns = deepcopy(self.uns)
        for k, df in adata.uns.items():
            if k.startswith("repguide_mask"):
                new_uns[k] = df.loc[guides_include, adata.var.rep.unique()]
            if not isinstance(df, pd.DataFrame):
                continue
            if "guide" in df.columns:
                if "allele" in df.columns:
                    key_col = ["guide", "allele"]
                elif "edit" in df.columns:
                    key_col = ["guide", "edit"]
                elif "aa_allele" in df.columns:
                    key_col = ["guide", "aa_allele"]
                else:
                    continue
                df_new = df.loc[df.guide.isin(guides_include), key_col + condit_include]
                # df_new = df_new.loc[df_new.loc[:, condit_include].sum(axis=1) > 0, :]
                new_uns[k] = df_new.reset_index(drop=True)
        adata.uns = new_uns
        return type(self).from_adata(adata)

    def remove_zero_allele_counts(
        self, key: Optional[Union[str, Iterable[str]]] = None
    ):
        condit_include = self.samples.index.tolist()

        def _remove_zero_count(key):
            df = self.uns[key]
            df = df.loc[df.loc[:, condit_include].sum(axis=1) > 0, :]
            self.uns[key] = df

        if key is None:
            for k in self.uns.keys():
                if "count" in k:
                    _remove_zero_count(k)
        elif isinstance(key, str):
            _remove_zero_count(key)
        elif isinstance(key, Iterable):
            for k in key:
                _remove_zero_count(k)

    def get_edit_mat_from_uns(
        self,
        ref_base: Optional[str] = None,
        alt_base: Optional[str] = None,
        match_target_position: Optional[bool] = None,
        rel_pos_start=0,
        rel_pos_end=np.Inf,
        rel_pos_is_reporter=False,
        target_pos_col="target_pos",
        edit_count_key="edit_counts",
    ):
        """
        Get the edit matrix from `.uns[edit_count_key]` to store the result in `.layers["edits"]`
        and returns the old `.layers["edits"]`.

        Args
        --
        ref_base: reference base of editing event to count. If not provided, default value in `.base_edited_from` is used.
        alt_base: alternate base of editing event to count. If not provided, default value in `.base_edited_to` is used.
        match_target_position: If `True`, edits with `.rel_pos` that matches `.guides[target_pos_col]` are counted.
        If `False`, edits with `.rel_pos` in range [rel_pos_start, rel_pos_end) is counted.
        rel_pos_start: Position from which edits will be counted when `match_target_position` is `False`.
        rel_pos_end: Position until where edits will be counted when `match_target_position` is `False`.
        rel_pos_is_reporter: `rel_pos_start` and `rel_pos_end` is relative to the reporter position. If `False`, those are treated to be relative of spacer position.
        target_pos_col: Column name in `.guides` DataFrame that has target position information when `match_target_position` is `True`.
        edit_count_key: Key of the edit counts DataFrame to be used to count the edits (`.uns[edit_count_key]`).
        """
        if ref_base is None:
            ref_base = self.base_edited_from
        if alt_base is None:
            alt_base = self.base_edited_to
        if match_target_position is None:
            match_target_position = not self.tiling
        if edit_count_key not in self.uns or len(self.uns[edit_count_key]) == 0:
            raise ValueError(
                "Edit count isn't calculated. "
                + "Call .get_edit_from_allele(allele_count_key, allele_key)"
            )
        edits = self.uns[edit_count_key].copy()
        if self.layers["edits"] is not None:
            old_edits = self.layers["edits"].copy()
            self.layers["edits"] = np.zeros_like(
                self.X,
            )
        else:
            old_edits = None
        edits["ref_base"] = edits.edit.map(lambda e: e.ref_base)
        edits["alt_base"] = edits.edit.map(lambda e: e.alt_base)
        edits = edits.loc[
            (edits.ref_base == ref_base) & (edits.alt_base == alt_base), :
        ].reset_index()
        guide_len = self.guides.sequence.map(len)
        guide_name_to_idx = self.guides.reset_index().reset_index().set_index("name")
        edits["guide_idx"] = guide_name_to_idx.loc[edits.guide, "index"].reset_index(
            drop=True
        )
        edits["guide_start_pos"] = (
            32 - 6 - guide_len[edits.guide_idx].reset_index(drop=True)
        )
        if not match_target_position:
            edits["rel_pos"] = edits.edit.map(lambda e: e.rel_pos)
            edits["in_rel_pos_range"] = (
                edits.rel_pos
                >= rel_pos_start + (0 if rel_pos_is_reporter else edits.guide_start_pos)
            ) & (
                edits.rel_pos
                < rel_pos_end + (0 if rel_pos_is_reporter else edits.guide_start_pos)
            )
            good_edits = edits.loc[
                edits.in_rel_pos_range, ["guide", "edit"] + self.samples.index.tolist()
            ]
        else:
            edits["rel_pos"] = edits.edit.map(lambda e: e.rel_pos)
            edits["target_pos_guides"] = self.guides.iloc[edits.guide_idx, :][
                target_pos_col
            ].reset_index(drop=True)
            edits["target_pos_matches"] = edits.rel_pos == edits.target_pos_guides
            good_edits = edits.loc[
                edits.target_pos_matches,
                ["guide", "edit"] + self.samples.index.tolist(),
            ]

        good_guide_idx = guide_name_to_idx.loc[good_edits.guide, "index"].astype(int)
        for gidx, eidx in zip(good_guide_idx, good_edits.index):
            self.layers["edits"][gidx, :] += good_edits.loc[
                eidx, self.samples.index.tolist()
            ].astype(int)
        print("New edit matrix saved in .layers['edits']. Returning old edits.")
        return old_edits

    def get_guide_edit_rate(
        self,
        normalize_by_editable_base: Optional[bool] = None,
        edited_base: Optional[str] = None,
        editable_base_start=3,
        editable_base_end=8,
        bcmatch_thres=1,
        prior_weight: float = None,
        return_result=False,
        count_layer="X_bcmatch",
        edit_layer="edits",
    ):
        """
        prior_weight:
        Considering the edit rate to have prior of beta distribution with mean 0.5,
        prior weight to use when calculating posterior edit rate.
        """
        if edited_base is None:
            edited_base = self.base_edited_from
        if normalize_by_editable_base is None:
            normalize_by_editable_base = self.tiling
        if self.layers[count_layer] is None or self.layers[edit_layer] is None:
            raise ValueError("edits or barcode matched guide counts not available.")
        num_targetable_sites = 1.0
        if normalize_by_editable_base:
            if edited_base not in ["A", "C", "T", "G"]:
                raise ValueError("Specify the correct edited_base")
            num_targetable_sites = self.guides.sequence.map(
                lambda s: s[editable_base_start:editable_base_end].count(edited_base)
            )
        bulk_idx = np.where(
            self.samples.reset_index()["index"].map(lambda s: "bulk" in s)
        )[0]

        if prior_weight is None:
            prior_weight = 1
        n_edits = self.layers[edit_layer][:, bulk_idx].sum(axis=1)
        n_counts = self.layers[count_layer][:, bulk_idx].sum(axis=1)
        edit_rate = (n_edits + prior_weight / 2) / (
            (n_counts * num_targetable_sites) + prior_weight / 2
        )
        edit_rate[n_counts < bcmatch_thres] = np.nan
        if normalize_by_editable_base:
            edit_rate[num_targetable_sites == 0] = np.nan
        if return_result:
            return edit_rate
        else:
            self.guides["edit_rate"] = edit_rate

    def get_edit_rate(
        self,
        normalize_by_editable_base=None,
        edited_base=None,
        editable_base_start=3,
        editable_base_end=8,
        bcmatch_thres=1,
        prior_weight: float = None,
        return_result=False,
        count_layer="X_bcmatch",
        edit_layer="edits",
    ):
        """
        prior_weight:
        Considering the edit rate to have prior of beta distribution with mean 0.5,
        prior weight to use when calculating posterior edit rate.
        """
        if normalize_by_editable_base is None:
            normalize_by_editable_base = self.tiling
        if edited_base is None:
            edited_base = self.base_edited_from
        if self.layers[count_layer] is None or self.layers[edit_layer] is None:
            raise ValueError("edits or barcode matched guide counts not available.")
        num_targetable_sites = 1.0
        if normalize_by_editable_base:
            if edited_base not in ["A", "C", "T", "G"]:
                raise ValueError("Specify the correct edited_base")
            num_targetable_sites = self.guides.sequence.map(
                lambda s: s[editable_base_start:editable_base_end].count(edited_base)
            )
        if prior_weight is None:
            prior_weight = 1
        n_edits = self.layers[edit_layer]
        n_counts = self.layers[count_layer].astype(float)
        n_counts[n_counts < bcmatch_thres] = np.nan
        edit_rate = n_edits / n_counts
        if normalize_by_editable_base:
            edit_rate[num_targetable_sites == 0, :] = np.nan
        if return_result:
            return edit_rate
        else:
            self.layers["edit_rate"] = edit_rate

    def get_edit_from_allele(
        self, allele_count_key="allele_counts", allele_key="allele", return_result=False
    ):
        if allele_count_key not in self.uns or len(self.uns[allele_count_key]) == 0:
            raise ValueError(f"No allele information stored: {self.uns}")

        df = self.uns[allele_count_key].copy()
        df = df.loc[df[allele_key].map(str) != "", :]
        df["edits"] = df[allele_key].map(lambda a: str(a).split(","))
        df = (
            df[["guide", "edits"] + self.samples.index.tolist()]
            .explode("edits")
            .groupby(["guide", "edits"])
            .sum()
        )
        df = df.reset_index().rename(columns={"edits": "edit"})

        def try_objectify(s):
            try:
                return Edit.from_str(s)
            except ValueError:
                return AminoAcidEdit.from_str(s)

        df["edit"] = df.edit.map(try_objectify)
        if return_result:
            return df
        self.uns["edit_counts"] = df

    def _get_allele_norm(self, allele_count_df=None, thres=10):
        """
        Get bcmatched count to normalize allele counts
        thres: normalizing count below this number would be disregarded.
        """
        if allele_count_df is None:
            allele_count_df = self.uns["allele_counts"]
        guide_to_idx = self.guides.reset_index().reset_index().set_index("name")
        guide_idx = guide_to_idx.loc[allele_count_df.guide, "index"]
        norm_counts = self.layers["X_bcmatch"][guide_idx.values.astype(int), :].copy()
        norm_counts[norm_counts < thres] = np.nan
        return norm_counts

    def get_normalized_allele_counts(self, allele_count_df=None, norm_thres=10):
        if allele_count_df is None:
            allele_count_df = self.uns["allele_counts"]
        alleles_norm = allele_count_df.copy()
        count_columns = self.samples.index.tolist()
        norm_counts = self._get_allele_norm(
            allele_count_df=allele_count_df, thres=norm_thres
        )
        alleles_norm.loc[:, count_columns] = (
            alleles_norm.loc[:, count_columns] / norm_counts
        )
        return alleles_norm

    def filter_allele_counts_by_pos(
        self,
        rel_pos_start: int = 0,
        rel_pos_end: int = 32,
        rel_pos_is_reporter: bool = True,
        allele_uns_key: str = "allele_counts",
        filter_rel_pos: bool = True,
        map_to_filtered: bool = True,
        distribute: bool = False,
        jaccard_threshold: float = 0.1,
    ):
        """
        Filter alleles based on barcode matched counts, allele counts,
        or proportion

        Args
        --
        map_to_filtered: Map allele to the closest filtered allele to preserve total allele count. Ignores the case where there is no alleles filtered.
        rel_pos_start: rel_pos to start including (inclusive)
        rel_pos_end: rel_pos to end including (exclusive)
        rel_pos_is_reporter: rel_pos_start and rel_pos_end is 0-based relative to reporter sequence start.
        jaccard_threshold (float) --
        """
        allele_count_df = self.uns[allele_uns_key].copy()
        if rel_pos_is_reporter:
            filtered_allele, filtered_edits = zip(
                *allele_count_df.allele.map(
                    lambda a: filter_allele_by_pos(
                        a, rel_pos_start, rel_pos_end, filter_rel_pos
                    )
                )
            )
        else:
            guide_start_pos = (
                32 - 6 - self.guides.loc[allele_count_df.guide, "guide_len"].values
            )
            filtered_allele, filtered_edits = zip(
                *[
                    filter_allele_by_pos(
                        allele_count_df.allele[i],
                        guide_start_pos[i] + rel_pos_start,
                        guide_start_pos[i] + rel_pos_end,
                        filter_rel_pos,
                    )
                    for i in range(len(allele_count_df))
                ]
            )
        allele_count_df.loc[:, "allele"] = filtered_allele
        # Hashing on Allele object messes up the order. Converting it to str and back to allele for groupby.
        allele_count_df["str_allele"] = allele_count_df.allele.map(str)
        allele_count_df = (
            allele_count_df.groupby(["guide", "str_allele"])
            .sum(numeric_only=True)
            .reset_index()
        )
        allele_count_df.insert(
            1, "allele", allele_count_df.str_allele.map(lambda s: Allele.from_str(s))
        )
        allele_count_df = allele_count_df.drop("str_allele", axis=1)
        print(
            f"{sum(filtered_edits)} edits filtered from {len(filtered_edits)} alleles."
        )
        if map_to_filtered:
            print("mapping filtered alleles ...")
            if distribute:
                allele_count_df = _distribute_alleles_to_filtered(
                    self.uns[allele_uns_key],
                    allele_count_df,
                    jaccard_threshold=jaccard_threshold,
                )
            else:
                allele_count_df = _map_alleles_to_filtered(
                    self.uns[allele_uns_key],
                    allele_count_df,
                    jaccard_threshold=jaccard_threshold,
                )
        return allele_count_df

    def filter_allele_counts_by_base(
        self,
        ref_base="A",
        alt_base="G",
        allele_uns_key="allele_counts",
        map_to_filtered=True,
        jaccard_threshold: float = 0.5,
    ):
        """
        Filter alleles based on base change.

        Keyword arguments:
        map_to_filtered -- Map allele to the closest filtered allele to preserve total allele count. Ignores the case where there is no alleles filtered.
        """
        allele_count_df = self.uns[allele_uns_key].copy()
        filtered_allele, filtered_edits = zip(
            *allele_count_df.allele.map(
                lambda a: filter_allele_by_base(
                    a, allowed_ref_base=ref_base, allowed_alt_base=alt_base
                )
            )
        )
        allele_count_df.loc[:, "allele"] = filtered_allele
        # Hashing on Allele object messes up the order. Converting it to str and back to allele for groupby.
        allele_count_df["str_allele"] = allele_count_df.allele.map(str)
        allele_count_df = (
            allele_count_df.groupby(["guide", "str_allele"])
            .sum(numeric_only=True)
            .reset_index()
        )
        allele_count_df.insert(
            1, "allele", allele_count_df.str_allele.map(lambda s: Allele.from_str(s))
        )
        allele_count_df = allele_count_df.drop("str_allele", axis=1)
        print(
            f"{sum(filtered_edits)} edits filtered from {len(filtered_edits)} alleles."
        )
        if map_to_filtered:
            allele_count_df = _map_alleles_to_filtered(
                self.uns[allele_uns_key],
                allele_count_df,
                jaccard_threshold=jaccard_threshold,
            )
        return allele_count_df

    def collapse_allele_by_target(
        self, ref_base, alt_base, target_pos_column="target_pos"
    ):
        if target_pos_column not in self.guides.columns:
            raise ValueError(
                f"The .guides have to have '{target_pos_column}' specifying the relative position of target edit."
            )
        df = self.uns["allele_counts"].copy().reset_index(drop=True)
        df["target_pos"] = self.guides.loc[df.guide, target_pos_column].reset_index(
            drop=True
        )
        df["has_target"] = df.apply(
            lambda row: row.allele.has_edit(ref_base, alt_base, rel_pos=row.target_pos),
            axis=1,
        )
        df["has_nontarget"] = df.apply(
            lambda row: row.allele.has_other_edit(
                ref_base, alt_base, rel_pos=row.target_pos
            ),
            axis=1,
        )
        df = df.drop("target_pos", axis=1)
        return df.groupby(["guide", "has_target", "has_nontarget"]).sum()

    def collapse_allele_by_nedit(self, ref_base, alt_base):
        df = self.uns["allele_counts"].copy().reset_index(drop=True)
        df["n_edits"] = df.allele.map(
            lambda a: sum(
                (
                    e.ref_base == ref_base
                    and e.alt_base == alt_base
                    and e.rel_pos >= 6
                    and e.rel_pos < 6 + 20
                )
                for e in a.edits
            )
        )
        df.loc[df.n_edits > 4, "n_edits"] = 4
        return df.groupby(["guide", "n_edits"]).sum()

    def translate_alleles(self):
        if self.uns["allele_counts"] is None:
            print("No allele information. Run crisrpep-count with -a option.")
            return
        self.uns["allele_counts"]["aa_allele"] = self.uns["allele_counts"].allele.map(
            lambda s: get_aa_alleles(s)
        )

    def log_norms(self, **kwargs):
        """Calculate log normalized guide counts and log normalized edit counts"""
        super().log_norm(**kwargs)
        super().log_norm(output_layer="lognorm_edits", read_count_layer="edits")

    def log_fold_changes(self, cond1, cond2, return_result=False):
        """Calculate log fold changes in normalized guide and edit counts."""
        guide_fc = super().log_fold_change(cond1, cond2, return_result=return_result)
        edit_fc = super().log_fold_change(
            cond1,
            cond2,
            lognorm_counts_key="lognorm_edits",
            out_guides_suffix="edit_lfc",
            return_result=return_result,
        )
        if return_result:
            return (guide_fc, edit_fc)

    def log_fold_change_aggregates(
        self,
        cond1,
        cond2,
        aggregate_condit="replicate",
        compare_condit="sort",
        aggregate_fn="median",
        return_result=False,
        keep_per_replicate=False,
    ):
        guide_fc_agg = super().log_fold_change_aggregate(
            cond1,
            cond2,
            lognorm_counts_key="lognorm_counts",
            aggregate_condit=aggregate_condit,
            compare_condit=compare_condit,
            out_guides_suffix="lfc",
            aggregate_fn=aggregate_fn,
            return_result=return_result,
            keep_per_replicate=keep_per_replicate,
        )
        edit_fc_agg = super().log_fold_change_aggregate(
            cond1,
            cond2,
            lognorm_counts_key="lognorm_edits",
            aggregate_condit=aggregate_condit,
            compare_condit=compare_condit,
            out_guides_suffix="edit_lfc",
            aggregate_fn=aggregate_fn,
            return_result=return_result,
            keep_per_replicate=keep_per_replicate,
        )
        if return_result:
            return (guide_fc_agg, edit_fc_agg)

    def allele_log_fold_change_aggregate(
        self,
        cond1,
        cond2,
        aggregate_condit="replicate",
        compare_condit="sort",
        aggregate_fn="median",
        return_result=False,
        keep_per_replicate=False,
    ):
        pass

    def write(self, out_path):
        """
        Write .h5ad
        """
        adata = self.copy()
        for k, v in adata.uns.items():
            if isinstance(v, pd.DataFrame) and len(v) > 0:
                if "edit" in v.columns and isinstance(
                    adata.uns[k]["edit"].iloc[0], Edit
                ):
                    adata.uns[k].edit = adata.uns[k].edit.map(str)
                for c in [colname for colname in v.columns if "allele" in colname]:
                    if isinstance(v[c].iloc[0], (Allele, CodingNoncodingAllele)):
                        adata.uns[k].loc[:, c] = adata.uns[k][c].map(str)
        super(ReporterScreen, adata).write(out_path)


def _convert_obj_column_to_str(df, obj_column):
    if obj_column not in df.columns:
        return df
    df = df.rename(columns={obj_column: "obj_col"})
    df[obj_column] = df["obj_col"].map(str)
    df = df.drop("obj_col", axis=1)
    return df


def concat(screens: Collection[ReporterScreen], *args, axis=1, **kwargs):
    """Concatenate multiple Screen objects."""
    # TODO: var/obs info not concated if doesn't overlap
    # TODO: concat 2 times: allele_counts merging doesn't work?
    if axis == 1 and not all(
        screen.guides.index.equals(screens[0].guides.index) for screen in screens
    ):
        raise ValueError("Guide index doesn't match.")

    adata = ad.concat(screens, *args, axis=axis, **kwargs)
    if axis == 1:
        adata.obs = screens[0].guides
    elif axis == 0:
        adata.var = screens[0].samples
    else:
        raise ValueError(f"axis must be 0 or 1. Provided of invalid axis {axis}.")
    keys = set(screens[0].uns.keys())
    for screen in screens:
        keys.intersection(screen.uns.keys())

    if axis == 0:
        for k in keys:
            if k in ["target_base_change", "tiling"]:
                adata.uns[k] = screens[0].uns[k]
                continue
            elif "edit" not in k and "allele" not in k:
                continue
            adata.uns[k] = pd.concat([screen.uns[k] for screen in screens])

    if axis == 1:
        # If combining multiple samples, edit/allele tables should be merged.
        for k in keys:
            if k in ["target_base_change", "tiling"]:
                adata.uns[k] = screens[0].uns[k]
                continue
            elif "edit" not in k and "allele" not in k:
                continue
            screen_uns_dfs = [screen.uns[k] for screen in screens]

            if max(len(screen_uns_df) for screen_uns_df in screen_uns_dfs) == 0:
                continue
            merge_on = screen[0].uns[k].columns[:2].tolist()
            try:
                edit_cls = type(screens[0].uns[k][merge_on[1]][0])
            except IndexError:
                print(merge_on)
                print(screen[0].uns[k])
                exit(1)
            screen_uns_dfs = [
                _convert_obj_column_to_str(df, merge_on[1]) for df in screen_uns_dfs
            ]

            for i, screen in enumerate(screens):
                if screen.uns[k][merge_on].duplicated().any():
                    raise ValueError(
                        f"{i}-th ReporterScreen .uns['{k}'] has duplicated row. Aborting."
                    )
                if i == 0:
                    adata.uns[k] = screen.uns[k]
                else:
                    adata.uns[k] = pd.merge(
                        adata.uns[k], screen.uns[k], on=merge_on, how="outer"
                    )
            adata.uns[k] = adata.uns[k].fillna(0)

            df = adata.uns[k].rename(columns={merge_on[1]: "_str"})
            df.insert(1, merge_on[1], df["_str"].map(lambda s: edit_cls.from_str(s)))
            adata.uns[k] = df.drop("_str", axis=1)

            float_col = adata.uns[k].select_dtypes(include=["float64"])
            for col in float_col.columns.values:
                adata.uns[k][col] = adata.uns[k][col].astype("int64")

    return ReporterScreen.from_adata(adata)


def read_h5ad(filename):
    adata = ad.read_h5ad(filename)
    return ReporterScreen.from_adata(adata)



from typing import Union, Sequence, Optional
import numpy as np
import pandas as pd
from scipy.stats import norm
from scipy.special import logit, expit
from statsmodels.stats.multitest import fdrcorrection


def get_fdr(mu_z, plot=False):
    p_dec = norm.cdf(mu_z)
    p_inc = norm.cdf(-mu_z)
    _, fdr_dec = fdrcorrection(p_dec)
    _, fdr_inc = fdrcorrection(p_inc)
    fdr = np.minimum(fdr_dec, fdr_inc)
    return (fdr_dec, fdr_inc, fdr)


def write_result_table(
    target_info_df: pd.DataFrame,
    param_hist_dict,
    model_label: str,
    prefix: str = "",
    suffix: str = "",
    write_fitted_eff: bool = True,
    guide_index: Optional[Sequence[str]] = None,
    guide_acc: Optional[Sequence] = None,
    return_result: bool = False,
) -> Union[pd.DataFrame, None]:
    """Combine target information and scores to write result table to a csv file or return it."""
    if param_hist_dict["params"]["mu_loc"].dim() == 2:
        mu = param_hist_dict["params"]["mu_loc"].detach()[:, 0].cpu().numpy()
        mu_sd = param_hist_dict["params"]["mu_scale"].detach()[:, 0].cpu().numpy()
        sd = param_hist_dict["params"]["sd_loc"].detach().exp()[:, 0].cpu().numpy()
    elif param_hist_dict["params"]["mu_loc"].dim() == 1:
        mu = param_hist_dict["params"]["mu_loc"].detach().cpu().numpy()
        mu_sd = param_hist_dict["params"]["mu_scale"].detach().cpu().numpy()
        sd = param_hist_dict["params"]["sd_loc"].detach().exp().cpu().numpy()
    else:
        raise ValueError(
            f'`mu_loc` has invalid shape of {param_hist_dict["params"]["mu_loc"].shape}'
        )
    mu_z = mu / mu_sd
    fdr_dec, fdr_inc, fdr = get_fdr(mu_z)
    fit_df = pd.DataFrame(
        {
            "mu": mu,
            "mu_sd": mu_sd,
            "mu_z": mu / mu_sd,
            "sd": sd,
            "fdr_dec": fdr_dec,
            "fdr_inc": fdr_inc,
            "fdr": fdr,
        }
    )
    if "negctrl" in param_hist_dict.keys():
        print("Normalizing with common negative control distribution")
        mu0 = param_hist_dict["negctrl"]["params"]["mu_loc"].detach().cpu().numpy()
        sd0 = (
            param_hist_dict["negctrl"]["params"]["sd_loc"].detach().exp().cpu().numpy()
        )
        print(f"Fitted mu0={mu0}, sd0={sd0}.")
        fit_df["mu_adj"] = (mu - mu0) / sd0
        fit_df["mu_sd_adj"] = mu_sd / sd0
        fit_df["mu_z_adj"] = fit_df.mu_adj / fit_df.mu_sd_adj
        fit_df["sd_adj"] = sd / sd0
        fdr_dec, fdr_inc, fdr = get_fdr(fit_df.mu_z_adj)
        fit_df["fdr_dec_adj"], fit_df["fdr_inc_adj"], fit_df["fdr_adj"] = (
            fdr_dec,
            fdr_inc,
            fdr,
        )
    fit_df = pd.concat(
        [target_info_df.reset_index(), fit_df.reset_index(drop=True)], axis=1
    )

    if write_fitted_eff or guide_acc is not None:
        if "alpha_pi" not in param_hist_dict["params"].keys():
            pi = 1.0
        else:
            a_fitted = param_hist_dict["params"]["alpha_pi"].detach().cpu().numpy()
            pi = a_fitted[..., 1:].sum(axis=1) / a_fitted.sum(axis=1)
        sgRNA_df = pd.DataFrame({"edit_eff": pi}, index=guide_index)
        if guide_acc is not None:
            sgRNA_df["accessibility"] = guide_acc
            sgRNA_df["scaled_edit_eff"] = _scale_pi(
                pi,
                guide_acc,
                fitted_noise_logit=param_hist_dict["params"]["noise_scale"]
                .detach()
                .cpu()
                .numpy(),
            )
        sgRNA_df.to_csv(f"{prefix}bean_sgRNA_result.{model_label}{suffix}.csv")

    if return_result:
        return fit_df
    fit_df.to_csv(f"{prefix}bean_element_result.{model_label}{suffix}.csv")


def _scale_edited_pi(
    pi: np.ndarray,
    guide_accessibility: np.ndarray,
    a: float = 0.2513,
    b: float = -1.9458,
):
    """Scale editied pi by its accessibility.

    Data fitted through relationship observed from data (updated 1/17/2023).
    Transformation derived from linear regression of log(endo/reporter) ~ a*log(atac_signal + 1)+b.
    If pi of multiple alleles are provided, pi is scaled so that total scaled pi wouldn't exceed 1.0.

    Args
    pi: Editing rate
    guide_accessibility: raw accessibility score
    """
    return pi * np.exp(b) * guide_accessibility**a


def _add_noise_to_pi(pi: np.ndarray, fitted_noise_logit: np.ndarray):
    logit_pi = logit(pi.clip(min=1e-3, max=1 - 1e-3)) + fitted_noise_logit
    return expit(logit_pi).clip(min=1e-3, max=1 - 1e-3)


def _scale_pi(pi: np.ndarray, guide_acc: np.ndarray, fitted_noise_logit=None):
    scaled_pi = _scale_edited_pi(pi, guide_acc)
    if fitted_noise_logit is None:
        return scaled_pi, None
    return _add_noise_to_pi(scaled_pi, fitted_noise_logit)



import numpy as np
import pandas as pd
from perturb_tools._qc.qc import get_outlier_guides


def get_outlier_guides_and_mask(
    bdata,
    condit_col: str,
    replicate_col: str = "replicate",
    mad_z_thres: float = 5,
    abs_RPM_thres: float = 10000,
):
    """Obtains outlier guides and per-(replicate, guide) mask to mask the outlier guides.
    For each experimental condition in `samples`, find outlier guides that shows extreme counts compared to guides.
    For each condition in `condit_col`, guides both with median absolute deviation z > `mad_z_thres` AND the RPM larger than `abs_RPM_thres` will be defined as outlier.

    Args
    --
    bdata: ReporterScreen object
    condit_col: column in bdata.samples with experimental samples
    replicate_col: column in bdata.samples with experimental replicate
    mad_z_thres: median absolute deviation threshold used to define outlier guides.
    abs_RPM_thres: RPM threshold value that will be used to define outlier guides.
    """
    outlier_guides = get_outlier_guides(bdata, condit_col, mad_z_thres, abs_RPM_thres)
    outlier_guides[replicate_col] = bdata.samples.loc[
        outlier_guides["sample"], replicate_col
    ].values
    mask = pd.DataFrame(
        index=bdata.guides.index, columns=bdata.samples[replicate_col].unique()
    ).fillna(1)
    for _, row in outlier_guides.iterrows():
        mask.loc[row[bdata.guides.index.name], row[replicate_col]] = 0
    return outlier_guides, mask



from __future__ import annotations
from typing import Iterable
import numpy as np
import re
from ..utils.arithmetric import jaccard


class Edit:
    reverse_map = {"A": "T", "C": "G", "T": "A", "G": "C", "-": "-"}
    strand_map = {"+": 1, "-": -1}

    def __init__(
        self,
        rel_pos: int,
        ref_base: chr,
        alt_base: chr,
        offset: int = None,
        strand: int = 1,
        unique_identifier=None,
    ):
        assert strand in [+1, -1]
        self.rel_pos = rel_pos
        self.ref_base = ref_base  # TODO make it ref / alt instead of ref_base and alt_base for AAEdit comp. or make abstract class
        self.alt_base = alt_base
        self.uid = unique_identifier
        if type(strand) == int:
            strand_to_symbol = {1: "+", -1: "-"}
            self.strand = strand_to_symbol[strand]
        else:
            assert strand in {"+", "-"}
            self.strand = strand
        self.pos = self.rel_pos if offset is None else offset + self.rel_pos * strand

    @classmethod
    def from_str(cls, edit_str):  # pos:strand:start>end
        if type(edit_str) is Edit:
            return edit_str
        if not cls.match_str(edit_str):
            raise ValueError(f"{edit_str} doesn't match with Edit string format.")
        uid = None
        if "!" in edit_str:
            uid, edit_str = edit_str.split("!")

        pos, rel_pos, strand, base_change = edit_str.split(":")
        pos = int(pos)
        rel_pos = int(rel_pos)
        assert strand in ["+", "-"]
        strand = cls.strand_map[strand]
        offset = pos - rel_pos * strand
        ref_base, alt_base = base_change.split(">")
        return cls(
            rel_pos,
            ref_base,
            alt_base,
            offset=offset,
            strand=strand,
            unique_identifier=uid,
        )

    @classmethod
    def match_str(cls, edit_str):
        if isinstance(edit_str, Edit):
            return True
        pattern = r"-?\d+:-?\d+:[+-]:[A-Z*-]>[A-Z*-]"
        pattern2 = r"[\w*]!-?\d+:-?\d+:[+-]:[A-Z*-]>[A-Z*-]"
        return re.fullmatch(pattern, edit_str) or re.fullmatch(pattern2, edit_str)

    def get_abs_edit(self):
        """
        Returns absolute edit representation regardless of the relative edit position by guide.
        """
        if self.strand == "-":
            ref_base = type(self).reverse_map[self.ref_base]
            alt_base = type(self).reverse_map[self.alt_base]
        else:
            ref_base = self.ref_base
            alt_base = self.alt_base
        if self.uid is not None:
            return f"{self.uid}!{int(self.rel_pos)}:{ref_base}>{alt_base}"
        return f"{int(self.pos)}:{ref_base}>{alt_base}"

    def set_uid(self, uid):
        if "!" in uid:
            raise ValueError("Cannot use special character `!` in uid.")
        self.uid = uid
        return self

    def get_abs_base_change(self):
        if self.strand == "-":
            ref_base = type(self).reverse_map[self.ref_base]
            alt_base = type(self).reverse_map[self.alt_base]
        else:
            ref_base = self.ref_base
            alt_base = self.alt_base
        return f"{ref_base}>{alt_base}"

    def get_base_change(self):
        ref_base = self.ref_base
        alt_base = self.alt_base
        return f"{ref_base}>{alt_base}"

    def __eq__(self, other):
        return self.__repr__() == other.__repr__()

    def __lt__(self, other):
        if isinstance(other, Edit) and self.pos != other.pos:
            return self.pos < other.pos
        return self.__repr__() < str(other)

    def __gt__(self, other):
        if isinstance(other, Edit) and self.pos != other.pos:
            return self.pos > other.pos
        return self.__repr__() > str(other)

    def __hash__(self):
        return hash(self.__repr__())

    def __repr__(self):
        if self.uid is None:
            return f"{int(self.pos)}:{int(self.rel_pos)}:{self.strand}:{self.ref_base}>{self.alt_base}"

        return f"{self.uid}!{int(self.pos)}:{int(self.rel_pos)}:{self.strand}:{self.ref_base}>{self.alt_base}"


class Allele:
    # pos, ref, alt
    def __init__(self, edits: Iterable[Edit] = None):
        self.edits = set() if edits is None else set(edits)

    @classmethod
    def from_str(cls, allele_str):  # pos:strand:start>end
        if type(allele_str) is Allele:
            return allele_str
        edits = set()
        try:
            for edit_str in allele_str.split(","):
                edit = Edit.from_str(edit_str)
                edits.add(edit)
        except ValueError:
            if allele_str.strip() == "":
                return cls(None)
        return cls(edits)

    @classmethod
    def match_str(cls, allele_str):
        if isinstance(allele_str, Allele):
            return True
        if allele_str == "":
            return True
        return all(map(Edit.match_str, allele_str.split(",")))

    def has_edit(self, ref_base, alt_base, pos=None, rel_pos=None):
        if not (pos is None) + (rel_pos is None):
            raise ValueError("Either pos or rel_pos should be specified")

        return any(
            e.ref_base == ref_base
            and e.alt_base == alt_base
            and (
                pos is not None and e.pos == pos or pos is None and e.rel_pos == rel_pos
            )
            for e in self.edits
        )

    def has_other_edit(self, ref_base, alt_base, pos=None, rel_pos=None):
        """
        Has other edit than specified.
        """
        if len(self.edits) == 0:
            return False
        if not (pos is None) + (rel_pos is None):
            raise ValueError("Either pos or rel_pos should be specified")
        for e in self.edits:
            if e.ref_base == ref_base and e.alt_base == alt_base:
                if (
                    pos is not None
                    and e.pos != pos
                    or pos is None
                    and e.rel_pos != rel_pos
                ):
                    return True
            else:
                return True
        return False

    def get_jaccard(self, other):
        return jaccard(set(map(str, self.edits)), set(map(str, other.edits)))

    def get_jaccards(self, allele_list: Iterable[Allele]):
        return np.array(list(map(lambda o: self.get_jaccard(o), allele_list)))

    def map_to_closest(
        self, allele_list, jaccard_threshold=0.5, merge_priority: np.ndarray = None
    ):
        """
        Arguments
        merge_priority -- Priority on which allele to merge if the jaccard index is the same.
        """
        if len(allele_list) == 0:
            return Allele()
        nt_jaccards = np.array(list(map(lambda o: self.get_jaccard(o), allele_list)))
        if not np.isnan(np.nanmax(nt_jaccards)):
            nt_max_idx = np.where(nt_jaccards == np.nanmax(nt_jaccards))[0]
            if len(nt_max_idx) > 0:
                if len(nt_max_idx) > 1 and merge_priority is not None:
                    if len(merge_priority) != len(allele_list):
                        raise ValueError(
                            f"merge_priority length {len(merge_priority)} is not the same as allele_list length {len(allele_list)}"
                        )
                    nt_max_idx = nt_max_idx[np.argmax(merge_priority[nt_max_idx])]
                else:
                    nt_max_idx = nt_max_idx[0]
                if nt_jaccards[nt_max_idx] > jaccard_threshold:
                    return allele_list[nt_max_idx.item()]
        return Allele()

    def __bool__(self):
        return len(self.edits) > 0

    def __len__(self):
        return len(self.edits)

    def __eq__(self, other):
        return self.__repr__() == other.__repr__()

    def __lt__(self, other):  # Implemented for pandas compatibility
        return len(self.edits) < len(other.edits)

    def __hash__(self):
        return hash(self.__repr__())

    def add(self, edit: Edit):
        self.edits.add(edit)  # TBD: adding to set?

    def update(self, edits: Iterable[Edit]):
        self.edits.update(edits)

    def __repr__(self):
        if len(self.edits) == 0:
            return ""
        list_edits = sorted(list(self.edits.copy()))
        list_edits = list(map(lambda s: str(s), list_edits))
        return ",".join(list_edits)



"""Calculate sample quality"""
from typing import Literal
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt

linestyles = ["solid", "dotted", "dashed", "dashdot"]


def plot_guide_edit_rates(bdata, ax=None, figsize=(5, 3), title="", n_bins=30):
    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)
    sns.histplot(bdata.guides.edit_rate, bins=n_bins)
    ax.set_title(title)
    ax.set_xlabel("Editing rate")
    return ax


def plot_sample_edit_rates(
    bdata, ax=None, figsize=(5, 3), title="", agg_method="median"
):
    set_sample_edit_rates(bdata, agg_method)
    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)
    for i, sample in enumerate(bdata.samples.index):
        sns.kdeplot(
            bdata.layers["edit_rate"][:, i],
            label=f"{sample}(median {np.nanmedian(bdata.layers['edit_rate'][:, i]):.2f})",
            linestyle=linestyles[(i // 10) % 4],
            clip=(0, 1),
        )
    sns.kdeplot(
        bdata.guides.edit_rate, linewidth=2, c="black", label="Bulk median", clip=(0, 1)
    )
    ax.set_xlabel("Editing rate")
    ax.set_title(title)
    ax.legend(bbox_to_anchor=(1, 1))
    return ax


def set_sample_edit_rates(
    bdata, agg_method: Literal["mean", "median"] = "median"
) -> None:
    if "edit_rate" not in bdata.layers or bdata.layers["edit_rate"].max() == 0:
        raise ValueError(
            "ReporterScreen object doesn't have .edit_rate. Calculate target site edits first using .get_edit_rate(...)"
        )
    if agg_method == "median":
        bdata.samples[f"{agg_method}_editing_rate"] = np.nanmedian(
            bdata.layers["edit_rate"], axis=0
        )
    if agg_method == "mean":
        bdata.samples[f"{agg_method}_editing_rate"] = np.nanmean(
            bdata.layers["edit_rate"], axis=0
        )



from copy import deepcopy
from functools import reduce
import multiprocessing
import numpy as np
from scipy.stats import fisher_exact
from tqdm.auto import tqdm
import pandas as pd
from scipy.special import softmax
from ..framework.Edit import Allele
from ..framework.AminoAcidEdit import CodingNoncodingAllele
from ._supporting_fn import map_alleles_to_filtered


def sum_column_groups(mat, column_index_list):
    cols = [
        mat[:, np.array(colidx)].sum(axis=1, keepdims=True)
        for colidx in column_index_list
    ]
    return np.concatenate(cols, axis=1)


def fisher_test_single_sample_multiproc(j):
    odds_ratio_series = pd.Series(dtype="float64").reindex_like(ctrl_edit)
    p_value_series = pd.Series(dtype="float64").reindex_like(ctrl_edit)
    for i in tqdm(range(len(sample_edit)), leave=False):
        if sample_edit.iloc[i, j] == 0:
            odds_ratio_series[i], p_value_series[i] = 0, 1
            continue
        fisher_tbl = [
            [sample_edit.iloc[i, j], guide_count_sample[i, j] - sample_edit.iloc[i, j]],
            [ctrl_edit.iloc[i], guide_count_ctrl[i] - ctrl_edit.iloc[i]],
        ]
        odds_ratio_series[i], p_value_series[i] = fisher_exact(
            fisher_tbl, alternative="greater"
        )
    return (odds_ratio_series, p_value_series)


def fisher_test_single_sample(
    j, sample_edit, ctrl_edit, guide_count_sample, guide_count_ctrl
):
    odds_ratio_series = pd.Series(dtype="float64").reindex_like(ctrl_edit)
    p_value_series = pd.Series(dtype="float64").reindex_like(ctrl_edit)
    for i in tqdm(range(len(sample_edit)), leave=False):
        if sample_edit.iloc[i, j] == 0:
            odds_ratio_series[i], p_value_series[i] = 0, 1
            continue
        fisher_tbl = [
            [sample_edit.iloc[i, j], guide_count_sample[i, j] - sample_edit.iloc[i, j]],
            [ctrl_edit.iloc[i], guide_count_ctrl[i] - ctrl_edit.iloc[i]],
        ]
        assert (np.array(fisher_tbl) >= 0).all(), fisher_tbl
        odds_ratio_series[i], p_value_series[i] = fisher_exact(
            fisher_tbl, alternative="greater"
        )
    return (odds_ratio_series, p_value_series)


def get_edit_significance_to_ctrl(
    sample_adata,
    ctrl_adata,
    aggregate_cond=None,
    run_parallel=False,
    allele_counts_key="allele_counts",
):
    """
    Calculate edit counts for all edits in sample_adata, regardless of the edit presence in the sample.
    """
    if "index" in sample_adata.samples.columns:
        sample_columns = sample_adata.samples["index"].tolist()
    else:
        sample_columns = sample_adata.samples.index.tolist()
    n_samples = len(sample_columns)
    sample_adata.uns["edit_counts"] = sample_adata.get_edit_from_allele(
        allele_count_key=allele_counts_key, return_result=True
    )
    ctrl_adata.uns["edit_counts"] = ctrl_adata.get_edit_from_allele(
        allele_count_key=allele_counts_key, return_result=True
    )
    edit_counts_ctrl = ctrl_adata.uns["edit_counts"]

    if aggregate_cond is not None:
        conds = sample_adata.samples.reset_index().groupby(aggregate_cond).groups
        edit_counts_raw = sample_adata.uns["edit_counts"][
            ["guide", "edit"] + sample_columns
        ].set_index(["guide", "edit"])
        edit_counts_sample = sample_adata.uns["edit_counts"][
            ["guide", "edit"]
        ].set_index(["guide", "edit"])
        for k, col_idx in conds.items():
            edit_counts_sample[k] = edit_counts_raw.iloc[
                :, np.array(col_idx.tolist())
            ].sum(axis=1)
        n_samples = len(conds.keys())
        sample_columns = conds.keys()
    else:
        edit_counts_sample = sample_adata.uns["edit_counts"][
            ["guide", "edit"] + sample_columns
        ]
    edit_counts_merged = pd.merge(
        edit_counts_sample, edit_counts_ctrl, on=["guide", "edit"], how="left"
    ).fillna(0)
    guide_count_ctrl = ctrl_adata._get_allele_norm(edit_counts_merged, thres=0)[:, 0]
    guide_count_sample = sample_adata._get_allele_norm(edit_counts_merged, thres=0)
    edit_counts_merged = edit_counts_merged.set_index(["guide", "edit"])
    if aggregate_cond is not None:
        guide_count_sample = sum_column_groups(guide_count_sample, conds.values())
    sample_edit = edit_counts_merged[sample_columns]
    ctrl_edit = edit_counts_merged.iloc[:, -1]

    print("Running Fisher's exact test to get significant edits compared to control...")

    def child_initialize2(
        _sample_edit, _ctrl_edit, _guide_count_sample, _guide_count_ctrl
    ):
        global sample_edit, ctrl_edit, guide_count_sample, guide_count_ctrl
        sample_edit = _sample_edit
        ctrl_edit = _ctrl_edit
        guide_count_sample = _guide_count_sample
        guide_count_ctrl = _guide_count_ctrl

    if run_parallel:
        with multiprocessing.Pool(
            min(n_samples, 30),
            initializer=child_initialize2,
            initargs=(sample_edit, ctrl_edit, guide_count_sample, guide_count_ctrl),
        ) as pool:
            odds_ratios, p_values = zip(
                *pool.map(fisher_test_single_sample_multiproc, range(n_samples))
            )
    else:
        odds_ratios, p_values = zip(
            *map(
                lambda i: fisher_test_single_sample(
                    i, sample_edit, ctrl_edit, guide_count_sample, guide_count_ctrl
                ),
                range(n_samples),
            )
        )
    p_value_tbl = pd.concat(p_values, axis=1)
    p_value_tbl.columns = sample_columns
    odds_ratio_tbl = pd.concat(odds_ratios, axis=1)
    odds_ratio_tbl.columns = sample_columns

    q_bonf_tbl = pd.DataFrame().reindex_like(p_value_tbl)
    for j in range(len(sample_columns)):
        q_bonf_tbl.iloc[:, j] = p_value_tbl.iloc[:, j] * (
            np.sum(edit_counts_sample[sample_columns].iloc[:, j] > 0)
            - np.sum(np.isnan(p_value_tbl.iloc[:, j]))
        )
    q_bonf_tbl[q_bonf_tbl > 1] = 1
    return (odds_ratio_tbl, q_bonf_tbl)


# https://stackoverflow.com/questions/8804830/python-multiprocessing-picklingerror-cant-pickle-type-function
def _filter_single_allele(allele: Allele, sig_edits):
    filtered_allele = deepcopy(allele)
    for edit in allele.edits:
        if edit not in sig_edits:
            filtered_allele.edits.remove(edit)
    return filtered_allele


def _row_filter_allele(row, sample_guide_to_sig_edit_dict):
    try:
        sig_edits = sample_guide_to_sig_edit_dict[row.guide]
    except KeyError:
        return Allele()
    return _filter_single_allele(row.allele, sig_edits)


def _filter_allele_sample(sample_allele_tbl, sample_guide_to_sig_edit_dict):
    sample_allele_tbl = sample_allele_tbl.reset_index()
    sample_allele_tbl["filtered_allele_str"] = sample_allele_tbl.apply(
        lambda row: _row_filter_allele(row, sample_guide_to_sig_edit_dict), axis=1
    ).map(str)
    sample_filtered_allele_tbl = (
        sample_allele_tbl.groupby(["guide", "filtered_allele_str"])
        .sum(numeric_only=True)
        .reset_index()
    )
    sample_filtered_allele_tbl = sample_filtered_allele_tbl.loc[
        sample_filtered_allele_tbl.filtered_allele_str != "", :
    ]
    return sample_filtered_allele_tbl.set_index(
        ["guide", "filtered_allele_str"], drop=True
    )


def _filter_allele_sample_multiproc(i):
    sample_allele_df = allele_df.iloc[:, i]
    sample_allele_df = sample_allele_df[sample_allele_df > 0]
    if filter_each_sample:
        sample_sig_edit = edit_significance_tbl.iloc[:, i] < q_thres
        if odds_ratio_tbl is not None and odds_ratio_thres is not None:
            sample_sig_edit = sample_sig_edit & (
                odds_ratio_tbl.iloc[:, i] > odds_ratio_thres
            )
    else:
        sample_sig_edit = (edit_significance_tbl < q_thres).any(axis=1)
        if odds_ratio_tbl is not None and odds_ratio_thres is not None:
            sample_sig_edit = sample_sig_edit & (odds_ratio_tbl > odds_ratio_thres).any(
                axis=1
            )
    sample_sig_edit = sample_sig_edit.loc[sample_sig_edit]
    sample_guide_to_sig_edit_dict = (
        sample_sig_edit.index.to_frame()
        .reset_index(drop=True)
        .groupby("guide")["edit"]
        .apply(list)
        .to_dict()
    )
    return _filter_allele_sample(sample_allele_df, sample_guide_to_sig_edit_dict)


def _filter_allele_sample_loop(
    i,
    allele_df,
    edit_significance_tbl,
    q_thres,
    filter_each_sample=False,
    odds_ratio_tbl=None,
    odds_ratio_thres=None,
):
    sample_allele_df = allele_df.iloc[:, i]
    sample_allele_df = sample_allele_df[sample_allele_df > 0]
    if filter_each_sample:
        sample_sig_edit = edit_significance_tbl.iloc[:, i] < q_thres
        if odds_ratio_tbl is not None and odds_ratio_thres is not None:
            sample_sig_edit = sample_sig_edit & (
                odds_ratio_tbl.iloc[:, i] > odds_ratio_thres
            )
    else:
        sample_sig_edit = (edit_significance_tbl < q_thres).any(axis=1)
        if odds_ratio_tbl is not None and odds_ratio_thres is not None:
            sample_sig_edit = sample_sig_edit & (odds_ratio_tbl > odds_ratio_thres).any(
                axis=1
            )
    sample_sig_edit = sample_sig_edit.loc[sample_sig_edit]
    sample_guide_to_sig_edit_dict = (
        sample_sig_edit.index.to_frame()
        .reset_index(drop=True)
        .groupby("guide")["edit"]
        .apply(list)
        .to_dict()
    )
    return _filter_allele_sample(sample_allele_df, sample_guide_to_sig_edit_dict)


def _filter_alleles(
    allele_df: pd.DataFrame,
    edit_significance_tbl: pd.DataFrame,
    q_thres: float,
    odds_ratio_tbl: pd.DataFrame = None,
    odds_ratio_thres: float = None,
    n_threads=None,
    filter_each_sample=False,
    run_parallel=False,
):
    """
    - args
        allele_df: DataFrame with guide(str), allele(Allele) and sample counts
        edit_significance_tbl:
        filter_each_sample (bool) : filter out edits that are insignificant in each of the sample.
        If False, preserve the edit if called significant in any of the sample.
    """
    if n_threads is None:
        n_threads = len(allele_df.columns)
    allele_df = allele_df.set_index(["guide", "allele"]).copy().fillna(0)

    def child_initialize(
        _allele_df,
        _edit_significance_tbl,
        _filter_each_sample,
        _q_thres,
        _odds_ratio_tbl=None,
        _odds_ratio_thres=None,
    ):
        # https://stackoverflow.com/questions/25825995/python-multiprocessing-only-one-process-is-running
        global allele_df, edit_significance_tbl, filter_each_sample, q_thres, odds_ratio_tbl, odds_ratio_thres
        allele_df = _allele_df
        edit_significance_tbl = _edit_significance_tbl
        filter_each_sample = _filter_each_sample
        q_thres = _q_thres
        odds_ratio_tbl = _odds_ratio_tbl
        odds_ratio_thres = _odds_ratio_thres

    if run_parallel:
        print("Running {} parallel processes to filter alleles...".format(n_threads))
        with multiprocessing.Pool(
            n_threads,
            initializer=child_initialize,
            initargs=(
                allele_df,
                edit_significance_tbl,
                filter_each_sample,
                q_thres,
                odds_ratio_tbl,
                odds_ratio_thres,
            ),
        ) as pool:
            filtered_allele_dfs = pool.map(
                _filter_allele_sample_multiproc, list(range(len(allele_df.columns)))
            )
    else:
        print("Filtering alleles...")
        filtered_allele_dfs = []
        for i in tqdm(range(len(allele_df.columns))):
            filtered_allele_dfs.append(
                _filter_allele_sample_loop(
                    i, allele_df, edit_significance_tbl, q_thres, filter_each_sample
                )
            )

    print("Done filtering alleles, merging the result...")

    try:
        filtered_alleles = reduce(
            lambda left, right: left.join(right, how="outer"), filtered_allele_dfs
        ).reset_index()
        filtered_alleles.insert(
            1,
            "allele",
            filtered_alleles.filtered_allele_str.map(lambda s: Allele.from_str(s)),
        )
        filtered_alleles = filtered_alleles.drop("filtered_allele_str", axis=1)
        return filtered_alleles
    except Exception as e:
        print(e)
        return filtered_allele_dfs


def filter_alleles(
    sample_adata,
    ctrl_adata,
    allele_counts_key="allele_counts",
    q_thres=0.05,
    odds_ratio_thres=2,
    aggregate_cond=None,
    filter_each_sample=False,
    edit_sig_tbl=None,
    n_threads=30,
    run_parallel=False,
    map_to_filtered=True,
):
    if aggregate_cond is not None:
        sample_tested = sample_adata.samples.groupby(aggregate_cond).ngroups
    elif not filter_each_sample:
        sample_tested = len(sample_adata.samples)
    else:
        sample_tested = 1

    q_thres /= sample_tested

    if edit_sig_tbl is None:
        odds_ratio_tbl, q_bonf_tbl = get_edit_significance_to_ctrl(
            sample_adata,
            ctrl_adata,
            aggregate_cond,
            run_parallel=run_parallel,
            allele_counts_key=allele_counts_key,
        )
        print("Done calculating significance.\n\n")
    else:
        q_bonf_tbl = edit_sig_tbl
        print("Using provided edit significance table.\n")
    print(
        f"Filtering alleles for those containing significant edits (q < {q_thres}) and odds ratio higher than {odds_ratio_thres}..."
    )
    filtered_alleles = _filter_alleles(
        sample_adata.uns[allele_counts_key],
        q_bonf_tbl,
        q_thres,
        odds_ratio_tbl,
        odds_ratio_thres,
        filter_each_sample=filter_each_sample,
        n_threads=n_threads,
        run_parallel=run_parallel,
    )
    if map_to_filtered:
        filtered_alleles = map_alleles_to_filtered(
            sample_adata.uns[allele_counts_key], filtered_alleles
        )
    print("Done!")
    return (q_bonf_tbl, filtered_alleles)


def filter_allele_prop(
    adata,
    allele_uns_key: str,
    allele_prop_thres: float = 0.05,
    allele_count_thres: int = 0,
    sample_prop_thres: float = 0.1,
    map_to_filtered=True,
    retain_max=True,
    allele_col: str = "allele",
    jaccard_threshold=0.5,
    aa_jaccard_threshold=0.5,
    nt_jaccard_threshold=0.5,
    distribute=False,
):
    """
    Filter allele based on the proportion of that allele among the guides.
    Arguments
    adata: ReporterScreen object (should be from bulk)
    -- allele_prop_thres: Proportion of allele among the barcode matched guide counts to filter for
    -- allele_count_thres: Read count of allele to filter for
    -- sample_prop_thres: Proportion of samples where allele proportion exceeds the allele_prop_thres.
    -- map_to_filtered: If True, map the allele counts that are filtered out to the closest and most abundant allele that is retained after filtering. If False, discard the read counts that are filtered out.
    -- retain_max: If True, in case there is no allele that is retained during the filtering step, retain one allele with maximum median allele proportion across samples.
    -- alelle_col: String key for the allele column in adata.uns[allele_uns_key]
    """
    alleles = adata.uns[allele_uns_key].copy()
    aa_prop = adata.get_normalized_allele_counts(alleles).set_index(
        ["guide", allele_col]
    )
    aa_prop_filtered = aa_prop.loc[
        (
            (np.nan_to_num(aa_prop.loc[:, adata.samples.index]) > allele_prop_thres)
            & (
                alleles.set_index(["guide", allele_col]).loc[:, adata.samples.index]
                >= allele_count_thres
            )
        ).mean(axis=1)
        >= sample_prop_thres,
        :,
    ]
    # if no allele is left for the guide, retain the maximum frequency allele

    if retain_max:
        filtered_guide = aa_prop_filtered.index.get_level_values("guide").unique()
        append_rows = []
        for guide in tqdm(aa_prop.index.get_level_values("guide").unique().tolist()):
            if guide not in filtered_guide:
                guide_df = aa_prop.loc[
                    aa_prop.index.get_level_values("guide") == guide,
                ]
                max_frequency_idx = np.argmax(guide_df.mean(axis=1))
                append_rows.append(guide_df.iloc[[max_frequency_idx], :])
        aa_prop_filtered = pd.concat([aa_prop_filtered] + append_rows, axis=0)

    alleles = alleles.set_index(["guide", allele_col])
    aa_filtered = alleles.loc[alleles.index.isin(aa_prop_filtered.index), :]
    alleles = alleles.reset_index()
    aa_filtered = aa_filtered.reset_index()
    if map_to_filtered:
        print("Mapping filtered alleles...")
        if distribute:
            aa_filtered = _distribute_alleles_to_filtered(
                alleles,
                aa_filtered,
                jaccard_threshold=jaccard_threshold,
                allele_col=allele_col,
            )
        else:
            aa_filtered = _map_alleles_to_filtered(
                alleles,
                aa_filtered,
                jaccard_threshold=jaccard_threshold,
                aa_jaccard_threshold=aa_jaccard_threshold,
                nt_jaccard_threshold=nt_jaccard_threshold,
                allele_col=allele_col,
            )
    return aa_filtered


def _map_alleles_to_filtered(
    raw_allele_counts: pd.DataFrame,
    filtered_allele_counts: pd.DataFrame,
    jaccard_threshold=0.5,
    aa_jaccard_threshold=0.5,
    nt_jaccard_threshold=0.5,
    allele_col="allele",
):
    """
    Map pre-filtering alleles to the closest post-filtering alleles and aggregate the read count.
    """
    mapped_allele_counts = []
    is_cn_allele = isinstance(
        raw_allele_counts.reset_index()[allele_col][0], CodingNoncodingAllele
    )
    for guide, guide_raw_counts in tqdm(
        raw_allele_counts.groupby("guide"),
        desc="Mapping alleles to closest filtered alleles",
    ):
        guide_filtered_allele_counts = filtered_allele_counts.loc[
            filtered_allele_counts.guide == guide, :
        ].set_index(allele_col)
        guide_filtered_alleles = guide_filtered_allele_counts.index.tolist()
        if len(guide_filtered_alleles) != 0:
            # prioritize allele with most mean counts
            merge_priority = guide_filtered_allele_counts.mean(
                axis=1, numeric_only=True
            )
            if is_cn_allele:
                guide_raw_counts["allele_mapped"] = guide_raw_counts[allele_col].map(
                    lambda allele: allele.map_to_closest(
                        guide_filtered_alleles,
                        aa_jaccard_threshold=aa_jaccard_threshold,
                        nt_jaccard_threshold=nt_jaccard_threshold,
                        merge_priority=merge_priority,
                    )
                )
            else:
                guide_raw_counts["allele_mapped"] = guide_raw_counts[allele_col].map(
                    lambda allele: allele.map_to_closest(
                        guide_filtered_alleles,
                        jaccard_threshold=jaccard_threshold,
                        merge_priority=merge_priority,
                    )
                )
            guide_raw_counts = (
                guide_raw_counts.drop(allele_col, axis=1)
                .rename(columns={"allele_mapped": allele_col})
                .groupby(["guide", allele_col])
                .sum()
            )

            mapped_allele_counts.append(guide_raw_counts)
    res = pd.concat(mapped_allele_counts).reset_index()
    res = res.loc[res[allele_col].map(bool), :]
    return res


def _distribute_alleles_to_filtered(
    raw_allele_counts: pd.DataFrame,
    filtered_allele_counts: pd.DataFrame,
    jaccard_threshold=0.1,
    allele_col="allele",
):
    """
    Distrubute pre-filtering allele counts to the post-filtering alleles based on
    similarity.
    """
    mapped_allele_counts = []
    is_cn_allele = isinstance(
        raw_allele_counts.reset_index()[allele_col][0], CodingNoncodingAllele
    )
    for guide, guide_filtered_counts in tqdm(
        filtered_allele_counts.groupby("guide"),
        desc="Mapping alleles to closest filtered alleles",
    ):
        guide_filtered_counts = guide_filtered_counts.set_index(["guide", allele_col])
        guide_raw_counts = raw_allele_counts.loc[
            raw_allele_counts.guide == guide, :
        ].set_index(["guide", allele_col])
        guide_filtered_alleles = guide_filtered_counts.index.get_level_values(
            allele_col
        ).tolist()
        res = guide_filtered_counts.values
        for i, (g, allele) in enumerate(guide_raw_counts.index):
            # for allelel in raw_counts, distribute based on similarity if not present in filtered_counts
            if allele in guide_filtered_alleles:
                continue
            if is_cn_allele:
                aa_jaccards, nt_jaccards = allele.get_jaccards(guide_filtered_alleles)
                jaccards = (aa_jaccards * 3 + nt_jaccards) / 4
            else:
                jaccards = allele.get_jaccards(guide_filtered_alleles)

            weights = np.zeros(len(jaccards))
            if not (jaccards > jaccard_threshold).any():
                # print(f"map {allele} to {guide_filtered_alleles} failed.")
                continue
            weights[jaccards > jaccard_threshold] = softmax(
                jaccards[jaccards > jaccard_threshold]
            )
            weights = np.minimum(weights, jaccards)
            dist_values = np.rint(
                weights[:, None] * guide_raw_counts.values[i, :][None, :]
            ).astype(int)
            if not (dist_values.sum(axis=0) <= guide_raw_counts.values[i, :]).all():
                pass  # rounding process can inflate total number of allele counts
            res += dist_values
        added_counts = pd.DataFrame(
            res,
            index=guide_filtered_counts.index,
            columns=guide_filtered_counts.columns,
        )
        mapped_allele_counts.append(added_counts)
    res = pd.concat(mapped_allele_counts).reset_index()
    res = res.loc[res[allele_col].map(bool), :]
    return res
